# pylint: skip-file
import pytest_asyncio
from ckms.core import Keychain

from cbra import Application
from cbra.session.const import SESSION_ENCRYPTION_KEY
from cbra.session.const import SESSION_SIGNING_KEY


@pytest_asyncio.fixture(scope='session') # type: ignore
async def keychain() -> Keychain:
    keychain = Keychain()
    keychain.configure({
        SESSION_SIGNING_KEY: {
            'provider': 'local',
            'kty': 'OKP',
            'use': 'sig',
            'key': {'path': 'pki/session-signing.key'}
        },
        SESSION_ENCRYPTION_KEY: {
            'provider': 'local',
            'public_kid': SESSION_ENCRYPTION_KEY,
            'kty': 'oct',
            'algorithm': 'A256GCM',
            'use': 'enc',
            'key': {'cek': b'0' * 32},
            'key_ops': ['decrypt', 'encrypt']
        }
    })
    await keychain
    return keychain


@pytest_asyncio.fixture # type: ignore
async def app(keychain: Keychain) -> Application:
    application = Application(
        keychain=keychain
    )
    await application.on_startup()
    return application


"""Declares :class:`ServerMetadata`."""
import typing

import pydantic


AUTH_METHODS = [
    "none",
    "client_secret_post",
    "client_secret_basic",
    "client_secret_jwt",
    "private_key_jwt",
    "tls_client_auth",
    "self_signed_tls_client_auth",
]


CONTENT_ENCRYPTION_ALGORITHMS = [
    "A128CBC-HS256",
    "A192CBC-HS384",
    "A256CBC-HS512",
    "A128GCM",
    "A192GCM",
    "A256GCM",
]

ENCRYPTION_ALGORITHMS = [
    "RSA1_5",
    "RSA-OAEP",
    "RSA-OAEP-256",
    "A128KW",
    "A192KW",
    "A256KW",
    "dir",
    "ECDH-ES",
    "ECDH-ES+A128KW",
    "ECDH-ES+A192KW",
    "ECDH-ES+A256KW",
    "A128GCMKW",
    "A192GCMKW",
    "A256GCMKW",
    "PBES2-HS256+A128KW",
    "PBES2-HS384+A192KW",
    "PBES2-HS512+A256KW",
]

SIGNATURE_ALGORITHMS = [
    "HS256",
    "HS384",
    "HS512",
    "RS256",
    "RS384",
    "RS512",
    "PS256",
    "PS384",
    "PS512",
    "ES256",
    "ES384",
    "ES512",
    "ES256K",
    "EdDSA"
]


class ServerMetadata(pydantic.BaseModel):
    __module__: str = 'cbra.ext.oauth2.types'

    issuer: str = pydantic.Field(
        default=None,
        title="Issuer",
        description="Authorization server's issuer identifier URL.",
        example="https://webid.unimatrixapis.com"
    )

    authorization_endpoint: str = pydantic.Field(
        default=None,
        title="Authorization endpoint",
        description="URL of the authorization server's authorization endpoint.",
        example="https://webid.unimatrixapis.com/authorize"
    )

    token_endpoint: str = pydantic.Field(
        default=None,
        title="Token endpoint",
        description="URL of the authorization server's token endpoint.",
        example="https://oauth2.unimatrixapis.com/token"
    )

    jwks_uri: typing.Optional[str] = pydantic.Field(
        default=None,
        title="JSON Web Key Set (JWKS) URI",
        description="URL of the authorization server's JWK Set document.",
        example="https://oauth2.unimatrixapis.com/jwks"
    )

    registration_endpoint: typing.Optional[str] = pydantic.Field(
        default=None,
        title="Registration endpoint",
        description=(
            "URL of the authorization server's **OAuth 2.0 Dynamic Client "
            "Registration Endpoint**."
        ),
        example="https://oauth2.unimatrixapis.com/register"
    )

    scopes_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported scopes",
        description=(
            "JSON array containing a list of the OAuth 2.0 `scope` values that "
            "this authorization server supports."
        ),
        example=["foo", "bar", "baz"]
    )

    response_types_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported response types",
        description=(
            "JSON array containing a list of the OAuth 2.0 `response_type` "
            "values that this authorization server supports."
        ),
        example=[
            "code",
            "code id_token",
            "code id_token token",
            "code token",
            "id_token",
            "id_token token",
            "none",
            "token"
        ]
    )

    response_modes_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported response modes",
        description=(
            "JSON array containing a list of the OAuth 2.0 `response_mode` "
            "values that this authorization server supports."
        ),
        example=[
            "fragment",
            "query",
        ]
    )

    grant_types_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported grant types",
        description=(
            "JSON array containing a list of the OAuth 2.0 `grant_types` "
            "values that this authorization server supports."
        ),
        example=[
            "authorization_code",
            "refresh_token",
            "client_credentials",
            "urn:ietf:params:oauth:grant-type:device_code",
            "urn:ietf:params:oauth:grant-type:jwt-bearer",
            "urn:ietf:params:oauth:grant-type:saml2-bearer",
            "urn:ietf:params:oauth:grant-type:token-exchange",
        ]
    )

    token_endpoint_auth_methods_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported client authentication methods",
        description=(
            "JSON array containing a list of client authentication methods "
            "supported by this token endpoint."
        ),
        example=AUTH_METHODS
    )

    token_endpoint_auth_signing_alg_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of the JWS signing algorithms "
            "supported by the token endpoint for the signature on the JWT "
            "used to authenticate the client at the token endpoint."
        ),
        example=SIGNATURE_ALGORITHMS
    )

    service_documentation: typing.Optional[str] = pydantic.Field(
        default=None,
        title="Documentation",
        description=(
            "URL of a page containing human-readable information that "
            "developers might want or need to know when using the "
            "authorization server."
        ),
        example="https://oauth2.unimatrixapis.com/docs"
    )

    ui_locales_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported locals for UI",
        description=(
            "Languages and scripts supported for the user interface, "
            "represented as a JSON array of language tag values from BCP 47."
        ),
        example=["nl-NL, be"]
    )

    op_policy_uri: typing.Optional[str]= pydantic.Field(
        default=None,
        title="Data policy URL",
        description=(
            "URL that the authorization server provides to the person "
            "registering the client to read about the authorization server's "
            "requirements on how the client can use the data provided by the "
            "authorization server."
        ),
        example="https://oauth2.unimatrixapis.com/privacy"
    )

    op_tos_uri: typing.Optional[str]= pydantic.Field(
        default=None,
        title="Terms of service URL",
        description=(
            "URL that the authorization server provides to the person "
            "registering the client to read about the authorization server's "
            "terms of service."
        ),
        example="https://oauth2.unimatrixapis.com/tos"
    )

    revocation_endpoint: typing.Optional[str] = pydantic.Field(
        default=None,
        title="Revocation endpoint",
        description="URL of the authorization server's revocation endpoint.",
        example="https://oauth2.unimatrixapis.com/revoke"
    )

    revocation_endpoint_auth_methods_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported authentication methods",
        description=(
            "JSON array containing a list of client authentication methods "
            "supported by this revocation endpoint."
        ),
        example=AUTH_METHODS
    )

    revocation_endpoint_auth_signing_alg_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of the JWS signing algorithms "
            "supported by the revocation endpoint for the signature on the JWT "
            "used to authenticate the client at the revocation endpoint."
        ),
        example=SIGNATURE_ALGORITHMS
    )

    introspection_endpoint: typing.Optional[str] = pydantic.Field(
        default=None,
        title="Introspection endpoint",
        description="URL of the authorization server's introspection endpoint.",
        example="https://oauth2.unimatrixapis.com/introspect"
    )

    introspection_endpoint_auth_methods_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported authentication methods",
        description=(
            "JSON array containing a list of client authentication methods "
            "supported by this introspection endpoint."
        ),
        example=AUTH_METHODS
    )

    introspection_endpoint_auth_signing_alg_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of the JWS signing algorithms "
            "supported by the introspection endpoint for the signature on the JWT "
            "used to authenticate the client at the introspection endpoint."
        ),
        example=SIGNATURE_ALGORITHMS
    )

    signed_metadata: typing.Optional[str] = pydantic.Field(
        default=None,
        title="Signed metadata",
        description=(
            "Signed JWT containing metadata values about the authorization "
            "server as claims."
        )
    )


    device_authorization_endpoint: typing.Optional[str] = pydantic.Field(
        default=None,
        title="Device authorization endpoint",
        description=(
            "URL of the authorization server's device authorization endpoint."
        ),
        example="https://oauth2.unimatrixapis.com/device"
    )

    tls_client_certificate_bound_access_tokens: typing.Optional[bool] = pydantic.Field(
        default=True,
        title="Supports mTLS certificate-bound access tokens",
        description=(
            "Indicates authorization server support for mutual-TLS client "
            "certificate-bound access tokens."
        ),
        example=True
    )

    mtls_endpoint_aliases: typing.Optional[typing.Dict[str, str]] = pydantic.Field(
        default={},
        title="Alternative mTLS endpoints",
        description=(
            "JSON object containing alternative authorization server "
            "endpoints, which a client intending to do mutual TLS will "
            "use in preference to the conventional endpoints."
        ),
        example={
            "authorize": "https://webid.unimatrixapis.com/authorize"
        }
    )

    nfv_token_signing_alg_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of the JWS signing algorithms "
            "supported by the server for signing the JWT used as NFV Token."
        ),
        example=SIGNATURE_ALGORITHMS
    )

    nfv_token_encryption_alg_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`alg` values) supported by the server to encode the JWT used as "
            "NFV Token."
        ),
        example=ENCRYPTION_ALGORITHMS
    )

    nfv_token_encryption_enc_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported content encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`enc` values) supported by the server to encode the JWT used as "
            "NFV Token."
        ),
        example=CONTENT_ENCRYPTION_ALGORITHMS
    )

    userinfo_endpoint: typing.Optional[str] = pydantic.Field(
        default=None,
        title="UserInfo endpoint",
        description="URL of the authorization servers' UserInfo Endpoint.",
        example="https://webid.unimatrixapis.com/userinfo"
    )

    acr_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported ACR",
        description=(
            "JSON array containing a list of the Authentication Context Class "
            "References that this authorization server supports."
        ),
        example=["a.b", "a"]
    )

    subject_types_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported subject types",
        description=(
            "JSON array containing a list of the Subject Identifier types that "
            "this authorization server supports"
        ),
        example=["pairwise", "public"]
    )

    id_token_signing_alg_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of the JWS signing algorithms "
            "supported by the server for signing the JWT used as ID Token."
        ),
        example=SIGNATURE_ALGORITHMS
    )

    id_token_encryption_alg_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`alg` values) supported by the server to encode the JWT used as "
            "ID Token."
        ),
        example=ENCRYPTION_ALGORITHMS
    )

    id_token_encryption_enc_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported content encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`enc` values) supported by the server to encode the JWT used as "
            "ID Token."
        ),
        example=CONTENT_ENCRYPTION_ALGORITHMS
    )

    userinfo_signing_alg_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of the JWS signing algorithms "
            "supported by the server for signing the JWT used as UserInfo Endpoint."
        ),
        example=SIGNATURE_ALGORITHMS
    )

    userinfo_encryption_alg_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`alg` values) supported by the server to encode the JWT used as "
            "UserInfo Endpoint."
        ),
        example=ENCRYPTION_ALGORITHMS
    )

    userinfo_encryption_enc_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported content encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`enc` values) supported by the server to encode the JWT used as "
            "UserInfo Endpoint."
        ),
        example=CONTENT_ENCRYPTION_ALGORITHMS
    )

    request_object_signing_alg_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of the JWS signing algorithms "
            "supported by the server for signing the JWT used as Request Object."
        ),
        example=SIGNATURE_ALGORITHMS
    )

    request_object_encryption_alg_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`alg` values) supported by the server to encode the JWT used as "
            "Request Object."
        ),
        example=ENCRYPTION_ALGORITHMS
    )

    request_object_encryption_enc_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported content encryption algorithms",
        description=(
            "JSON array containing a list of the JWE encryption algorithms "
            "(`enc` values) supported by the server to encode the JWT used as "
            "Request Object."
        ),
        example=CONTENT_ENCRYPTION_ALGORITHMS
    )

    display_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported display modes",
        description=(
            "JSON array containing a list of the `display` parameter values "
            "that the OpenID Provider supports."
        ),
        example=["page", "popup", "touch", "wap"]
    )

    claim_types_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported Claims Types",
        description=(
            "JSON array containing a list of the Claims Types "
            "that the OpenID Provider supports."
        )
    )

    claims_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported Claims Types",
        description=(
            "JSON array containing a list of the Claim Names of the Claims "
            "that the OpenID Provider MAY be able to supply values for."
        )
    )

    claims_locales_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported claim locales",
        description=(
            "Languages and scripts supported for values in Claims being "
            "returned, represented as a JSON array of BCP 47."
        ),
        example=["nl-NL", "be"]
    )

    claims_parameter_supported: typing.Optional[bool] = pydantic.Field(
        default=False,
        title="Supports `claims` parameter?",
        description=(
            "Boolean value specifying whether the OP supports use of the "
            "`claims` parameter."
        ),
        example=True
    )

    request_parameter_supported: typing.Optional[bool] = pydantic.Field(
        default=False,
        title="Supports `request` parameter?",
        description=(
            "Boolean value specifying whether the OP supports use of the "
            "`request` parameter."
        ),
        example=True
    )

    request_uri_parameter_supported: typing.Optional[bool] = pydantic.Field(
        default=False,
        title="Supports `request_uri` parameter?",
        description=(
            "Boolean value specifying whether the OP supports use of the "
            "`request_uri` parameter."
        ),
        example=True
    )

    require_request_uri_registration: typing.Optional[bool] = pydantic.Field(
        default=True,
        title="Requires pre-regiration?",
        description=(
            "Boolean value specifying whether the OP requires any `request_uri` "
            "values used to be pre-registered."
        ),
        example=True
    )

    require_signed_request_object: typing.Optional[bool] = pydantic.Field(
        default=True,
        title="Requires pre-regiration?",
        description=(
            "Indicates where authorization request needs to be protected as "
            "**Request Object** and provided through either `request` or "
            "`request_uri` parameter."
        ),
        example=True
    )

    pushed_authorization_request_endpoint: typing.Optional[str] = pydantic.Field(
        default=None,
        title="Pushed Authorization Request (PAR) endpoint",
        description=(
            "URL of the authorization server's pushed authorization request "
            "endpoint."
        ),
        example="https://oauth2.unimatrixapis.com/par"
    )

    require_pushed_authorization_requests: typing.Optional[bool] = pydantic.Field(
        default=False,
        title="Requires PAR?",
        description=(
            "Indicates whether the authorization server accepts authorization "
            "requests only via PAR."
        ),
        example=False
    )

    introspection_signing_alg_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported signature algorithms",
        description=(
            "JSON array containing a list of algorithms supported by the "
            "authorization server for introspection response signing."
        ),
        example=SIGNATURE_ALGORITHMS
    )

    introspection_encryption_alg_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported encryption algorithms",
        description=(
            "JSON array containing a list of algorithms supported by the "
            "authorization server for introspection response content key "
            "encryption (`alg` value)."
        ),
        example=ENCRYPTION_ALGORITHMS
    )

    introspection_encryption_enc_values_supported: typing.Optional[typing.List[str]] = pydantic.Field(
        default=[],
        title="Supported content encryption algorithms",
        description=(
            "JSON array containing a list of algorithms supported by the "
            "authorization server for introspection response content "
            "encryption (`enc` value)."
        ),
        example=CONTENT_ENCRYPTION_ALGORITHMS
    )

    authorization_response_iss_parameter_supported: typing.Optional[bool] = pydantic.Field(
        default=True,
        title="Supports `iss` parameter in authorization response?",
        description=(
            "Boolean value indicating whether the authorization server "
            "provides the `iss` parameter in the authorization response."
        ),
        example=True
    )

    # TODO: These are non-standard properties and we must submit an RFC
    # to register these at IANA.
    token_endpoint_auth_encryption_alg_values_supported: typing.Optional[typing.List[str]] = []
    token_endpoint_auth_encryption_enc_values_supported: typing.Optional[typing.List[str]] = []
    assertion_signing_alg_values_supported: typing.Optional[typing.List[str]] = []
    assertion_encryption_alg_values_supported: typing.Optional[typing.List[str]] = []
    assertion_encryption_enc_values_supported: typing.Optional[typing.List[str]] = []
    required_encrypted_token_endpoint_auth: typing.Optional[bool] = False
    require_encrypted_assertion: typing.Optional[bool] = False


import typing

import pydantic

import cbra


class GrandParent(pydantic.BaseModel):
    foo: str


class Parent(pydantic.BaseModel):
    parent: GrandParent


class Child(pydantic.BaseModel):
    parent: Parent



class ExampleEndpoint(cbra.Endpoint):
    require_authentication = False
    model=Child

    async def handle( # type: ignore
        self,
        dto: typing.Optional[typing.Dict[str, typing.Any]] = None
    ) -> typing.Dict[str, typing.Any]: # type: ignore
        """Example endpoint."""
        return {
            "message": "Hello world!",
            "body": dto if dto else None
        }


app = cbra.Application()
app.add(ExampleEndpoint, base_path='/get', method="GET")
app.add(ExampleEndpoint, base_path='/post', method="POST")


if __name__ == '__main__':
    cbra.run('__main__:app', reload=True)



"""Declares :class:`RedirectURL`."""
import operator
import urllib.parse
from typing import Any
from typing import Generator

from ..exceptions import MissingRedirectURL
from ..exceptions import RedirectForbidden


class RedirectURL:
    """Represents the redirect URI for the **Authorization Endpoint**."""
    __module__: str = 'cbra.ext.oauth2.types'
    params: dict[str, Any]
    url: urllib.parse.ParseResult

    @classmethod
    def __get_validators__(cls) -> Generator[Any, None, None]:
        yield cls.validate

    @classmethod
    def __modify_schema__(
        cls,
        field_schema: dict[str, Any]
    ) -> None: # pragma: no cover
        field_schema.update({
            "type": "str"
        })

    @classmethod
    def validate(
        cls,
        value: str | None
    ) -> "RedirectURL":
        """Parse `value` into a :class:`urllib.parse.ParseResult`
        instance and verify that the URL is served over HTTPS and
        is redirectable.
        """
        if not value:
            return cls(None)
        if isinstance(value, cls):
            return value
        assert isinstance(value, str) # nosec
        p = urllib.parse.urlparse(value)
        if p.scheme != "https" and p.hostname != '127.0.0.1':
            raise RedirectForbidden(
                error_description="Redirect URL must serve over HTTPS."
            )
        if p.fragment:
            raise RedirectForbidden(
                error_description="Redirect URL must not contain a fragment."
            )
        if not p.netloc:
            raise RedirectForbidden(
                error_description=f"Unable to redirect to {value}."
            )
        return cls(p)

    def __init__(self, url: urllib.parse.ParseResult | None = None):
        self.url = url
        self.params = {}
        if self.url is not None and self.url.query:
            self.params = urllib.parse.parse_qs(self.url.query)

    def authorize(self, **params: str | None) -> str:
        """Authorize the redirect to the URL specified by the :class:`RedirectURI`
        and return a new instance with the query parameters.
        """
        if self.url is None:
            raise MissingRedirectURL
        params = {
            **self.params,
            **{k: v for k, v in params.items() if v is not None}
        }
        parts = list(self.url)
        parts[4] = urllib.parse.urlencode(params, True)
        return urllib.parse.urlunparse(parts)

    def is_loopback(self) -> bool:
        """Return a boolean indicating if the URL is the loopback
        interface.
        """
        return self.url.hostname == "127.0.0.1"

    def __bool__(self) -> bool:
        return self.url is not None

    def __str__(self) -> str:
        return urllib.parse.urlunparse(self.url) if self.url else ''

    def __eq__(self, url: object) -> bool:
        x = urllib.parse.urlparse(str(self))
        y = urllib.parse.urlparse(str(url))
        if "127.0.0.1" in {x.hostname, y.hostname}:
            is_equal = operator.eq(
                (x.scheme, x.hostname, x.path, x.query),
                (y.scheme, y.hostname, y.path, y.query),
            )
        else:
            is_equal = operator.eq(
                (x.scheme, x.netloc, x.path, x.query),
                (y.scheme, y.netloc, y.path, y.query),
            )
        return is_equal

    def __repr__(self) -> str:
        return f"<RedirectURI: {self}>"


"""Declares :class:`BaseGrant`."""
from cbra.ext.oauth2.types.isubject import ISubject
import pydantic

from .clientassertiontype import ClientAssertionType
from .fields import ResourceField
from .granttype import GrantType
from .iclient import IClient
from .stringorlist import StringOrList


class BaseGrant(pydantic.BaseModel):
    __module__: str = 'cbra.ext.oauth2.types'
    _client: IClient | None = pydantic.PrivateAttr(default=None)

    grant_type: GrantType
    client_id: str | None = pydantic.Field(
        default=None,
        title="Client ID",
        description=(
            "Identifies the client that is requesting a token on the behalf of "
            "the resource owner. If the client is public or does not use the "
            "`Authorization` header to authenticate, this parameter is "
            "mandatory."
        )
    )

    client_secret: str | None = pydantic.Field(
        default=None,
        title="Client secret",
        description=(
            "Required if the client is confidential and the request does not "
            "use the `Authorization` header or other supported scheme to "
            "authenticate, else this parameter must be not present. It may "
            "also be omitted if the client secret is an empty string."
        )
    )

    client_assertion_type: ClientAssertionType | None = pydantic.Field(
        default=None,
        title="Client assertion type",
        description=(
            "The authorization server supports the following client assertions "
            "to establish the identity of confidential OAuth 2.0 clients:\n\n"
            "`urn:ietf:params:oauth:client-assertion-type:jwt-bearer`\n\n"
            "Clients that have registered a public key sign a JWT using that "
            "key. The Client authenticates in accordance with **JSON Web Token "
            "(JWT) Profile for OAuth 2.0 Client Authentication and Authorization "
            "Grants** and **Assertion Framework for OAuth 2.0 Client Authentication "
            "and Authorization Grants**. The JWT must contain the following "
            "required Claim Values and may contain the following optional Claim "
            "Values:\n\n"
            "- `iss` - (Required) Issuer. This MUST contain the ``client_id`` of "
            "the OAuth Client.\n"
            "- `sub` - (Required) Issuer. This MUST contain the ``client_id`` of "
            "the OAuth Client.\n"
            "- `aud` - (Required) Value that identifies the authorization server "
            "as an intended audience. The authorization server must verify that "
            "it is an intended audience for the token. The Audience must be the "
            "URL of the authorization server's **Token Endpoint**.\n"
            "- `jti` - (Required) A unique identifier for the token, which can "
            "be used to prevent reuse of the token. These tokens must only be "
            "used once.\n"
            "- `exp` - (Required) Expiration time on or after which the ID "
            "Token must not be accepted for processing.\n"
            "- `iat` - (Optional) Time at which the JWT was issued."
        )
    )

    client_assertion: str | None = pydantic.Field(
        default=None,
        title="Client assertion",
        description=(
            "The client assertion used to authenticate a confidential OAuth 2.0 "
            "client. See `client_assertion_type` for the supported protocols and "
            "formats."
        )
    )

    resource: StringOrList = ResourceField()

    def get_audience(self) -> str | list[str] | None:
        """Return the audience to use when issueing a JWT access token
        as a result of the request.
        """
        return list(sorted(self.resource or []))

    def get_client(self) -> IClient:
        assert self._client is not None # nosec
        return self._client

    def has_client_assertion(self) -> bool:
        """Return a boolean indicating if the client authenticates itself
        using a client assertion.
        """
        return bool(self.client_assertion_type)\
            and bool(self.client_assertion)\
            and bool(self.client_id)

    def set_client(self, client: IClient) -> None:
        self._client = client

    def validate_grant(self, client: IClient, subject: ISubject | None) -> None:
        pass


# pylint: skip-file
# Copyright (C) 2022 Cochise Ruhulessin <cochiseruhulessin@gmail.com>
# 
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
import pytest
from httpx import AsyncClient


@pytest.mark.asyncio
async def test_authenticate_with_token(
    id_token: str,
    client: AsyncClient
):
    response = await client.post( # type: ignore
        url='/export',
        headers={'Authorization': f'Bearer {id_token}'}
    )
    assert response.status_code == 200


"""Declares :class:`ISubjectRepository`."""
from typing import Any

from ..exceptions import SubjectDoesNotExist # type: ignore
from .isubject import ISubject


class ISubjectRepository:
    __module__: str = 'cbra.ext.oauth2.types'
    SubjectDoesNotExist: type[BaseException] = SubjectDoesNotExist

    @classmethod
    def new(cls, **kwargs: Any) -> type['ISubjectRepository']:
        return type(cls.__name__, (cls,), kwargs)

    async def get(
        self,
        client: Any,
        subject_id: Any
    ) -> ISubject | None:
        """Return a :class:`ISubject` implementation for the client specified
        by `client_id` and `subject_id`. If `client_id` is ``None``, no properties
        specific to a client/subject relationship are loaded onto the resulting
        :class:`ISubject` instance.
        """
        raise NotImplementedError

    async def persist(self, client: Any, obj: Any | Any) -> ISubject:
        raise NotImplementedError


import fastapi

import cbra
from cbra.session import CookieSession


app = cbra.Application(
    docs_url='/'
)
app.keychain.register({
    'session-signing-key': {
        'provider': "local",
        'kty': "OKP",
        'alg': "EdDSA",
        'crv': 'Ed448',
        'use': 'sig'
    },
    'session-encryption-key': {
        'provider': "local",
        'public_kid': 'session-encryption-key',
        'kty': "oct",
        'alg': "dir",
        'use': 'enc',
        'length': 32
    }
})


class SessionEndpoint(cbra.Endpoint):
    document: bool = True
    method: str = "GET"
    session: CookieSession

    async def handle(self):
        await self.session
        return self.session.claims


@app.post('/session')
async def set(
    response: fastapi.Response,
    session: CookieSession = fastapi.Depends(),
    key: str = fastapi.Query('foo'),
    value: str = fastapi.Query(...)
):
    await session.set(key, value)
    await session.add_to_response(response=response)


app.add(SessionEndpoint, base_path='/session')
if __name__ == '__main__':
    cbra.run('__main__:app', reload=True)


# pylint: skip-file
from .base import BaseParser
from .base import IParser
from .form import FormParser
from .json import JSONParser
from .jwt import JWTParser
from .null import NullParser
from .yaml import YAMLParser


__all__ = [
    'BaseParser',
    'FormParser',
    'IParser',
    'JSONParser',
    'JWTParser',
    'NullParser',
    'YAMLParser',
]



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Declares :class:`WebhookEndpoint`."""
import asyncio
import functools
import warnings
from typing import cast
from typing import Any
from typing import Awaitable
from typing import NoReturn
from typing import TypeAlias

import fastapi
import pydantic
from httpx import AsyncClient
from httpx import BasicAuth

from cbra.ext import webhook
from cbra.parsers import JSONParser
from cbra.types import IParser
from . import models


class WebhookEndpoint(webhook.WebhookEndpoint):
    __module__: str = 'cbra.ext.picqer'
    description: str = (
        "Receive a Picqer event and queue it for upstream "
        "processing.\n\n"
        "Refer to https://picqer.com/en/api for additional "
        "documentation on the Picqer API."
    )
    model: TypeAlias = models.EventType
    parsers: list[type[IParser]] = [JSONParser]
    summary: str = "Picqer"
    signature: bytes | None = None

    @classmethod
    async def create_webhooks(
        cls,
        *,
        hooks: list[dict[str, Any]],
        secret: str,
        base_url: str,
        email: str,
        api_key: str,
        **kwargs: Any
    ) -> list[dict[str, Any]]: # pragma: no cover
        """Ensure that the given list of webhooks `hooks` exists at the
        Picqer instance identified by `base_url`.
        """
        params = {
            'auth': BasicAuth(username=api_key, password='x'),
            'base_url': base_url,
            'headers': {
                'User-Agent': f"MyPicqerClient (picqer.com/api - {email})"
            }
        }
        futures: list[Awaitable[dict[str, Any]]] = []
        async with AsyncClient(**params) as client:
            for hook in hooks:
                futures.append(
                    cls._get_or_create_webhook(client, hook, secret)
                )
            await asyncio.gather(*futures)

        return hooks

    @staticmethod
    async def _get_or_create_webhook(
        client: AsyncClient,
        hook: dict[str, Any],
        secret: str
    ) -> dict[str, Any]: # pragma: no cover
        response = await client.post( # type: ignore
            url='/api/v1/hooks',
            json=hook
        )
        if response.status_code not in {201, 400}:
            response.raise_for_status()
        if response.status_code == 400:
            dto = response.json()
            if dto.get('error_message') != "Error: Hook already exists":
                response.raise_for_status()
        if response.status_code == 201:
            remote = response.json()
            hook['idhook'] = remote['idhook']
        return hook

    def __init__(
        self,
        signature: str | None = fastapi.Header(
            default=None,
            title="Signature",
            alias='X-Picqer-Signature',
            description=(
                "A Base64-encoded Hash-based Message Authentication Code (HMAC) "
                "of the request body, that was created using a pre-shared secret."
            )
        )
    ):
        if signature is not None:
            self.signature = cast(bytes, self.decode_signature(signature))

    async def authenticate( # type: ignore
        self,
        dto: models.EventType
    ) -> NoReturn | None:
        """Verifies that the `X-WC-Signature` header was created using the
        preshared key.
        """
        await self.verify_request(
            signature=self.signature or b'',
            secret=self.get_hmac_secret()
        )

    @functools.singledispatchmethod
    async def get_object(
        self,
        dto: models.EventType
    ) -> dict[str, Any] | pydantic.BaseModel:
        """Return the object on which the event occurred."""
        raise NotImplementedError

    @functools.singledispatchmethod
    async def on_event(
        self,
        dto: models.EventType
    ) -> dict[str, Any] | None: # pragma: no cover
        """Handles the incoming event message or emit a warning if there is
        no handler defined.
        """
        return await self.sink(dto)

    @on_event.register
    async def on_order_closed(
        self,
        dto: models.OrderClosed
    ) -> dict[str, Any] | None:
        """Invoked when an ``orders.closed`` event is received."""
        return await self.sink(dto)

    @on_event.register
    async def on_free_stock_changed(
        self,
        dto: models.ProductFreeStockChanged
    ) -> dict[str, Any] | None:
        """Invoked when a ``products.free_stock_changed`` event is received."""
        return await self.sink(dto)

    async def sink(self, dto: models.EventType, **params: Any) -> dict[str, Any]:
        warnings.warn(
            f"{type(self).__name__} did not specify a handler method for "
            f"event type {dto.event} ({type(dto).__name__})"
        )
        return {
            'status': 'dropped',
            **params
        }


# type: ignore
"""Declares various utility functions."""
import functools
import typing


def dispatcher(
    func: typing.Callable[
        ...,
        typing.Coroutine[typing.Any, typing.Any, typing.Any]
    ]
) -> typing.Coroutine[typing.Any, typing.Any, typing.Any]:
    """Decorate a function to parse incoming requests and determine
    the appropriate event types.
    """
    handlers: dict[str, typing.Any] = {}

    @functools.wraps(func)
    async def dispatch(self, *args, **kwargs):
        event_name, params = await func(self, *args, **kwargs)
        handler =  handlers.get(event_name)
        if handler is None:
            raise AttributeError(f"No handler for {event_name}")
        return await handler(self, params)

    def register_handler(event_name):
        """Registers a handler with the given `event_name`."""
        def decorator_factory(func):
            handlers[event_name] = func
            return func
        return decorator_factory

    dispatch.on = register_handler
    dispatch.handlers = handlers
    return dispatch


class ClassPropertyDescriptor:

    @classmethod
    def new(cls, func):
        if not isinstance(func, (classmethod, staticmethod)):
            func = classmethod(func)
        return cls(func)

    def __init__(self, fget, fset=None):
        self.fget = fget
        self.fset = fset

    def __get__(self, obj, cls=None) -> typing.Any:
        if cls is None:
            cls = type(obj)
        return self.fget.__get__(obj, cls)()

    def __set__(self, obj, value) -> typing.NoReturn:
        if not self.fset:
            raise AttributeError("can't set attribute")
        type_ = type(obj)
        return self.fset.__get__(obj, type_)(value)

    def setter(self, func) -> typing.Any:
        if not isinstance(func, (classmethod, staticmethod)):
            func = classmethod(func)
        self.fset = func
        return self


classproperty = ClassPropertyDescriptor.new


"""Declares a basic session management framework."""
import datetime
import logging
from typing import Any
from typing import Generator

import fastapi
import cryptography.exceptions
from ckms.jose import PayloadCodec
from ckms.types import ClaimSet
from ckms.utils import b64encode_str
from unimatrix.exceptions import CanonicalException

from cbra.conf import settings # type: ignore
from cbra.params import ServerCodec
from cbra.params import CurrentServer
from .basesession import BaseSession
from .const import SESSION_COOKIE_NAME
from .const import SESSION_ENCRYPTION_KEYS
from .const import SESSION_SIGNING_KEYS


class SessionClaims(dict[str, Any]):

    def __init__(self, session: 'CookieSession', *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.session = session

    def update(self, *args: Any, **kwargs: Any):
        self.session.dirty = True
        return super().update(*args, **kwargs)



class CookieSession(BaseSession):
    """Represents a session that a user agent has with the
    authorization server.
    """
    audience: set[str] = set()
    domain: str | None = None
    issuer: str
    logger: logging.Logger = logging.getLogger('uvicorn')
    token: str | None
    token_type: str = "jwt+session"
    readonly: bool = False
    samesite: str = 'strict'

    def __init__(self,
        request: fastapi.Request,
        issuer: str = CurrentServer,
        codec: PayloadCodec = ServerCodec,
        token: str = fastapi.Cookie(
            default=None,
            alias=SESSION_COOKIE_NAME,
            title="Session",
            description=(
                "Identifies the current session associated to the user agent."
            )
        )
    ):
        self.claims = self.awaiting
        self.codec = codec
        self.dirty = set()
        self.issuer = issuer
        self.request = request
        self.token = token
        if request is not None:
            request.scope['session'] = self

    async def add_to_response(self, response: fastapi.Response) -> None:
        """Update the response with the modified session, if it was
        modified during request handling.
        """
        if self.readonly:
            return
        if not self.dirty and not self.created:
            return
        claims = self.claims if self.claims != self.awaiting else {}
        assert isinstance(claims, dict)
        response.set_cookie(
            key=SESSION_COOKIE_NAME,
            value=b64encode_str(
                await self.codec.encode(
                    payload=ClaimSet.parse_obj({
                        **claims,
                        'iss': self.issuer,
                        'aud': self.audience or self.issuer
                    }),
                    signers=SESSION_SIGNING_KEYS,
                    encrypters=SESSION_ENCRYPTION_KEYS,
                    content_type=self.token_type,
                    allow_compact=False,
                    allow_flattened=False
                )
            ),
            max_age=86000 * 365,
            secure=True,
            httponly=True,
            samesite=self.samesite,
            domain=self.domain,
            path=self.path
        )

    async def _deserialize(self) -> None:
        now = int(datetime.datetime.utcnow().timestamp())
        if self.token is None:
            self.created = True
            self.claims = {'iat': now}
        elif self.claims == self.awaiting:
            assert self.token is not None
            try:
                _, claims = await self.codec.jwt(
                    token=str.encode(self.token, 'ascii'),
                    accept="jwt+session"
                )
                claims.verify(
                    audience=self.audience or {self.issuer}
                )
                self.claims = SessionClaims(self, claims.dict())
            except (CanonicalException, cryptography.exceptions.InvalidTag) as exception:
                # The session was tampered with or we rotated the session key.
                self.logger.critical(
                    "Caught %s while decoding session cookie. "
                    "Session was probably tampered with.",
                    type(exception).__name__
                )
                self.created = True
                self.claims = {'iat': now}
            except Exception as exception:
                self.logger.exception("Caught fatal %s", type(exception).__name__)
                self.created = True
                self.claims = {'iat': now}

    async def get(self, key: str) -> Any:
        """Retrieve a key from the session, or ``None`` if it does
        not exist.
        """
        await self # type: ignore
        assert isinstance(self.claims, dict), repr(self.claims)
        return self.claims.get(key) # type: ignore

    async def set(self, key: str, value: Any) -> None:
        """Set a key to the session."""
        if key in self.reserved_keys:
            raise ValueError(f"Reserved session key: {key}")
        await self # type: ignore
        assert isinstance(self.claims, dict)
        if value != self.claims.get(key): # type: ignore
            self.dirty.add(key)
            self.claims[key] = value

    def __await__(self) -> Generator[Any, None, None]:
        return self._deserialize().__await__()



"""Declares :class:`IClientRepository`."""
import abc
from typing import Any

from ..exceptions import ClientDoesNotExist
from .configurable import Configurable
from .iclient import IClient


class IClientRepository(Configurable, metaclass=abc.ABCMeta):
    """Specifies the interface for :term:`IClient` repository
    implementations.
    """
    __module__: str = 'cbra.ext.oauth2.types'
    DoesNotExist: type[BaseException] = ClientDoesNotExist
    ClientDoesNotExist: type[BaseException] = DoesNotExist
    kwargs_name: str = 'clients'
    settings_key: str = 'OAUTH_CLIENTS'

    @classmethod
    def new(cls, **kwargs: Any) -> type['IClientRepository']:
        return type(cls.__name__, (cls,), kwargs)

    @abc.abstractmethod
    async def exists(self, client_id: str) -> bool:
        """Retur a boolean indicating if the client exists."""
        raise NotImplementedError

    @abc.abstractmethod
    async def get(self, client_id: str) -> IClient:
        """Lookup an OAuth 2.0 client using its client identifier."""
        raise NotImplementedError


import cbra
from cbra import Application
from cbra.ext.google import AortaEndpoint
from cbra.ext.google import CloudSchedulerEndpoint
from cbra.ext.google import EventarcEndpoint
from cbra.ext.google import GoogleEndpoint
from cbra.ext.google.models import MessagePublished


class EventarcMessageReceiver(EventarcEndpoint):
    pass


class ExportDailyStockList(CloudSchedulerEndpoint):
    summary: str = "Trigger stocklist export"
    response_description: str = "The stocklist is queued for export."

    async def handle(self):
        pass


class ExampleGoogleEndpoint(GoogleEndpoint):
    method: str = 'POST'

    async def handle(self):
        return {"email": self.principal.email if self.principal else "none"}


def get_asgi_application():
    app = Application()
    app.add(ExampleGoogleEndpoint, base_path='/')
    app.add(ExportDailyStockList, base_path='/export')
    app.add(AortaEndpoint, base_path='/aorta')
    app.add(EventarcMessageReceiver, base_path='/eventarc')
    return app


app = get_asgi_application()


@app.get('/foo')
def test(body: MessagePublished):
    pass


if __name__ == '__main__':
    cbra.run('__main__:app', reload=True)



"""Declares :class:`ISubject`."""
from typing import Any
from typing import Coroutine

from ckms.core.models import JSONWebSignature
from ckms.types import JSONWebKey


class ISubject:
    """Specifies the interface to which all subject implementations
    must provide.
    """
    __module__: str = 'cbra.ext.oauth2.types'

    #: The subject identifier. The type and semantics are implementation specific,
    #: but the concrete implementation must ensure that :attr:`sub` is JSON
    #: serializable.
    sub: int | str

    #: Identifies the client in which the context subject was
    #: instantiated, or ``None``.
    client_id: int | str | None = None

    #: Used for testing purposes only.
    #client_scope: defaultdict[str, set[str]]

    def allows_scope(self, scope: set[str]) -> bool:
        """Return a boolean indicating that the requested scope `scope`
        was allowed by the subject.
        """
        raise NotImplementedError

    def register_public_key(self, key: JSONWebKey) -> None:
        """Register a public key for the :class:`ISubject`."""
        raise NotImplementedError

    def verify(
        self,
        jws: JSONWebSignature
    ) -> bool | Coroutine[Any, Any, bool]:
        """Verifies a JSON Web Signature (JWS) using the keys that were
        previously registered by the :term:`Subject`.
        """
        raise NotImplementedError


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Declares :class:`IParser`."""
from typing import Any

from .request import Request


class IParser:
    __module__: str = 'cbra.types'
    encoding: str | None
    media_type: str

    @classmethod
    def openapi_example(cls, schema: dict[str, Any]) -> str:
        return ""

    def __init__(self,
        encoding: str | None = None
    ):
        self.encoding = encoding

    async def parse(self,
        request: Request,
        media_type: str | None = None,
        parser_context: dict[str, Any] | None = None
    ) -> Any:
        raise NotImplementedError


"""Declares :class:`IUpstreamProvider`."""
import fastapi

from .authorizationrequestparameters import AuthorizationRequestParameters
from .istorage import IStorage


class IUpstreamProvider:
    __module__: str = 'cbra.ext.oauth2.types'
    name: str

    async def create_redirect(
        self,
        request: fastapi.Request,
        dto: AuthorizationRequestParameters
    ) -> str:
        """Create the redirect URI that initiates the authentication and/or
        authorization flow at the upstream identity provider.
        """
        raise NotImplementedError

    async def get_authorization_request(
        self,
        storage: IStorage,
        request: fastapi.Request
    ) -> AuthorizationRequestParameters:
        """Return an authorization request based on the return parameters
        provided by the upstream identity provider.
        """
        raise NotImplementedError

    def get_authorization_request_id(self, request: fastapi.Request) -> str:
        """Discover the authorization request identifier based on the
        return parameters sent by the upstream provider.
        """
        raise NotImplementedError

    def get_client_redirect(
        self,
        request: fastapi.Request,
        dto: AuthorizationRequestParameters
    ) -> str:
        """Get the client redirect URL."""
        raise NotImplementedError

    async def on_return(
        self,
        storage: IStorage,
        request: fastapi.Request
    ) -> fastapi.responses.RedirectResponse:
        """Invoked when the upstream identity provider returns to the
        authorization server.
        """
        raise NotImplementedError

    async def process_response(
        self,
        request: fastapi.Request,
        params: AuthorizationRequestParameters
    ) -> None:
        """Handles the response data from the upstream provider."""
        raise NotImplementedError

    def reverse(self, request: fastapi.Request) -> str:
        """Get the return URL to which the upstream identity
        provider must redirect.
        """
        raise NotImplementedError



import asyncio
from cbra.ext.picqer import WebhookEndpoint


BASE_URL = "https://molano.picqer.com"
EMAIL = "sales@molano.nl"
API_KEY = "eJ8PQRiGdHyCygYcrhUVwDJNhz8PjXMqy9jF3Nwomj7xdHUr"
HOOKS = [
    {
        'name': "urn:webhook:picqer:products.stock_changed:uplink.molano.nl",
        'event': "products.stock_changed",
        'address': "https://uplink.molano.nl/log-request"
    },
    {
        'name': "urn:webhook:picqer:orders.completed:uplink.molano.nl",
        'event': "orders.completed",
        'address': "https://uplink.molano.nl/log-request"
    },
]


async def main():
    await WebhookEndpoint.create_webhooks(
        hooks=HOOKS,
        secret='foo',
        email=EMAIL,
        api_key=API_KEY,
        base_url=BASE_URL
    )


if __name__ == '__main__':
    asyncio.run(main())



"""Declares :class:`MemorySubjectRepository`."""
import typing

from .types import ISubject
from .types import ISubjectRepository


class MemorySubjectRepository(ISubjectRepository):
    """A :class:`ISubjectRepository` implementation that used local
    memory for storage. Mainly for testing purposes.
    """
    __module__: str = 'cbra.ext.oauth2'
    subjects: typing.Dict[typing.Union[int, str], ISubject] = {}

    @classmethod
    def add(cls, subject: ISubject) -> None:
        """Add a subject to the repository."""
        cls.subjects[subject.sub] = subject

    @classmethod
    def clear(cls) -> None:
        """Clears the repository of all data."""
        cls.subjects = {}

    @classmethod
    def configure(cls, subjects: typing.Dict[str, ISubject]) -> None:
        """Load a mapping of subject identifiers to subjects."""
        cls.subjects.update(subjects) # type: ignore

    def __init__(self):
        self.subjects = MemorySubjectRepository.subjects

    async def get(
        self,
        subject_id: typing.Union[int, str],
        client_id: typing.Optional[str] = None
    ) -> ISubject:
        try:
            subject = self.subjects[subject_id]
            subject.client_id = client_id
            return subject
        except KeyError:
            raise self.SubjectDoesNotExist


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# pylint: skip-file
# type: ignore
from typing import Any

from .basemessagetest import BaseMessageTest


class TestProductDelete(BaseMessageTest):
    event_type: str = 'deleted'
    resource_type: str = 'product'

    def get_request_params(self) -> dict[str, Any]:
        return {
            'headers': {
                'content-type': "application/json",
                'x-wc-webhook-delivery-id': "427d9a8e8269f1e7cbe7b26eacee4b4e",
                'x-wc-webhook-event': "deleted",
                'x-wc-webhook-id': "59",
                'x-wc-webhook-resource': "product",
                'x-wc-webhook-signature': "UYySC2m92uOtsOxltbamISzY3UU3Ns4DdlCm1HpEXyY=",
                'x-wc-webhook-source': "https://test-wordpress.molano.nl/",
                'x-wc-webhook-topic': "product.deleted",
            },
            'content': b'{"id":74}'
        }


"""Declares :class:`AuthorizationEndpoint`."""
import typing

import fastapi

from cbra.types import IPrincipal
from .authorizationrequestclient import AuthorizationRequestClient
from .endpoint import Endpoint
from .exceptions import CrossOriginNotAllowed
from .exceptions import Error
from .exceptions import LoginRequired
from .params import DownstreamProvider
from .params import LocalIssuer
from .params import Server
from .params import SubjectRepository
from .params import TransientStorage
from .params import UpstreamProvider
from .requestobjectdecoder import RequestObjectDecoder
from .rfc9068principal import RFC9068Principal
from .types import AuthorizationRequest
from .types import AuthorizationRequestParameters
from .types import IAuthorizationServer
from .types import IAuthorizeEndpoint
from .types import IClient
from .types import IStorage
from .types import ISubject
from .types import ISubjectRepository
from .types import IUpstreamProvider
from .types import JAR


class AuthorizationEndpoint(Endpoint, IAuthorizeEndpoint):
    __module__: str = 'cbra.ext.oauth2'
    summary: str = "Authorization Endpoint"
    decoder: RequestObjectDecoder
    description: str = (
        "The **Authorization Endpoint** is used to interact with the resource "
        "owner and obtain an authorization grant."
    )
    login_endpoint: str | None
    login_url: str | None
    methods: set[str] = {"GET"}
    query_model: type[AuthorizationRequestParameters] = AuthorizationRequestParameters
    require_authentication: bool = False
    options_description: str = (
        "Communicates the allowed methods and CORS options for "
        "the **Authorization Request** endpoint."
    )
    principal_factory: typing.Callable[..., IPrincipal] = RFC9068Principal(
        auto_error=False
    )
    provider: IUpstreamProvider | None

    def __init__(
        self,
        decoder: RequestObjectDecoder = fastapi.Depends(),
        issuer: str = LocalIssuer,
        server: IAuthorizationServer = Server,
        storage: IStorage = TransientStorage,
        subjects: ISubjectRepository = SubjectRepository,
        provider: IUpstreamProvider | None = UpstreamProvider
    ):
        self.decoder = decoder
        self.issuer = issuer
        self.login_url = server.login_url
        self.login_endpoint = server.login_endpoint
        self.provider = provider
        self.storage = storage
        self.subjects = subjects

    @staticmethod
    async def callback(
        request: fastapi.Request,
        provider: IUpstreamProvider = DownstreamProvider,
        storage: IStorage = TransientStorage
    ) -> fastapi.Response:
        """Return path for upstream identity providers."""
        return await provider.on_return(storage, request)

    async def add_to_client(
        self,
        client: IClient,
        principal: IPrincipal
    ) -> ISubject:
        """Add the principal given its subject identifier to the client. This
        method is invoked when handling an authorization request for a client
        with which a subject has never authenticated before.
        """
        return await self.subjects.persist(client, principal)

    async def create_redirect(
        self,
        dto: AuthorizationRequest
    ) -> str:
        """Return the redirect URI for a previously validated
        :class:`AuthorizationRequest`.
        """
        if self.provider is not None:
            return await self.create_upstream_redirect(dto=dto)

        redirect_uri = None
        assert dto.redirect_uri is not None # nosec
        assert bool(dto.authorization_code) # nosec
        redirect_uri = dto.redirect_uri.authorize(
            code=dto.authorization_code,
            iss=self.issuer,
            state=dto.state
        )
        assert redirect_uri is not None # nosec
        return redirect_uri

    async def create_upstream_redirect(
        self,
        dto: AuthorizationRequestParameters
    ) -> str:
        """Return a string containing the redirection endpoint at the
        upstream identity provider.
        """
        assert self.provider is not None # nosec
        return await self.provider.create_redirect(self.request, dto)

    async def enforce_cors_policy( # type: ignore
        self,
        params: AuthorizationRequestParameters = fastapi.Depends(),
        client: IClient = AuthorizationRequestClient,
        origin: str | None = fastapi.Header(
            default=None,
            alias='Origin'
        )
    ):
        if not client.allows_origin(origin):
            raise CrossOriginNotAllowed(
                redirect_uri=client.get_redirect_url(
                    url=params.redirect_uri,
                    fatal=False
                ),
                state=params.state
            )

    async def get_subject(
        self,
        client: IClient,
        principal: IPrincipal,
        dto: AuthorizationRequest
    ) -> ISubject | None:
        """Return a :class:`~cbra.ext.oauth2.types.ISubject` instance
        representing the currently authenticated principal.
        """
        subject = await dto.get_subject(client, self.subjects)
        if subject is None:
            # There was no credential attached to the authorization request
            # that could identify the subject.
            subject = await self.subjects.get(
                client=client,
                subject_id=principal.sub
            ) or await self.add_to_client(client, principal)
        assert subject is not None
        if subject and principal and (subject.sub != principal.sub):
            raise Error(
                error="invalid_request",
                error_description=(
                    "The authenticated subject does not match the subject "
                    "for which the authorization request was created."
                ),
                mode='client'
            )
        return subject

    async def resolve(
        self,
        client: IClient,
        params: AuthorizationRequestParameters
    ) -> AuthorizationRequest:
        """Inspect the authorization request and determine if it
        needs to be retrieved from the storage or an external source,
        if the `request` or `request_uri` parameters are provided.
        Otherwise return the request.
        """
        if params.request and params.request_uri:
            raise Error(
                error_description=(
                    "An authorization request can not include both the "
                    "\"request\" and \"request_uri\" parameters."
                )
            )
        if params.is_object():
            # A Request Object as specified in RFC 9101
            assert params.request is not None
            claims = await self.decoder.decode_request_object(params.request)
            return AuthorizationRequest.fromparams(JAR(**claims))

        if not params.is_reference():
            return AuthorizationRequest.fromparams(params)
        if params.is_external():
            raise Error(
                error="invalid_request_uri",
                error_description=(
                    "The application does not allow retrieving authorization "
                    "request parameters from the uri that was provided with "
                    "the \"request_uri\" parameter."
                )
            )

        assert params.request_id is not None # nosec
        return await self.storage.get_authorization_request(params.request_id)

    async def handle(
        self,
        client: IClient = AuthorizationRequestClient,
        params: AuthorizationRequestParameters = fastapi.Depends()
    ):
        """Initiates the **Authorization Code Flow**."""
        dto = await self.resolve(client, params)
        if not self.principal.is_authenticated():
            return await self.on_login_required(dto)
        subject = await self.get_subject(
            client=client,
            principal=self.principal,
            dto=dto
        )
        assert subject is not None # nosec

        # Check if the request parameters are valid.
        await self.validate_request(
            client=client,
            subject=subject,
            dto=dto
        )

        # Check if the subject allows the scope requested. If the subject
        # does not allow the scope requested, redirect to the consent_url
        # or present a page when the subject can allow the scope.
        if not subject.allows_scope(dto.scope):
            return self.redirect_consent(dto)

        redirect_uri = await self.create_redirect(dto=dto)

        assert dto.sub is not None
        await self.persist(dto)
        return fastapi.responses.RedirectResponse(
            url=redirect_uri,
            status_code=303
        )

    async def on_login_required(self, dto: AuthorizationRequest):
        """Raises an exception indicating that authentication is required."""
        if dto.is_openid() and not dto.can_interact():
            raise Error(
                error="login_required",
                error_description=(
                    "Provide credentials using the designated mechanisms."
                ),
                redirect_uri=dto.redirect_uri,
                mode='redirect'
            )

        if self.login_endpoint:
            raise NotImplementedError

        if self.login_url:
            raise LoginRequired(
                redirect_to=self.login_url,
                params={'next': str(self.request.url)}
            )

    async def persist(
        self,
        dto: AuthorizationRequest
    ) -> None:
        """Inspect the request parameters and persist the appropriate
        objects to the transient storage.
        """
        await self.storage.persist(dto)
        if dto.needs_authorization_code():
            await self.storage.persist(dto.get_authorization_code())

    def redirect_consent(self, dto: AuthorizationRequest) -> fastapi.Response:
        """Create a response that redirects the user-agent to a
        page where the resource owner can grant consent to the
        client.
        """
        if dto.is_openid() and not dto.can_interact():
            raise Error(
                error="consent_required",
                error_description=(
                    "The resource owner has not granted consent to this client "
                    "for the requested scope, and the authorization server "
                    "was instructed not to prompt the resource owner with an "
                    "interface to obtain consent.\n\n"
                    "This indicates a misconfiguration. Contact your system "
                    "administrator or support department for further "
                    "information."
                ),
                redirect_uri=dto.redirect_uri
            )
        raise NotImplementedError
        
    async def validate_request(
        self,
        client: IClient,
        subject: ISubject,
        dto: AuthorizationRequest
    ) -> None:
        """Validates an OAuth 2.0 authorization request."""
        await dto.validate_request_parameters(client, subject)
        dto.sub = typing.cast(str, subject.sub)


"""Declares :class:`RequestQuery`."""
import inspect
import typing
from inspect import Parameter
from inspect import Signature

import fastapi
import fastapi.params

from cbra.ext.ioc import Parameterless
from .namedquery import NamedQuery
from .pathfield import PathField
from .queryrunner import QueryRunner
from .requestquerymetaclass import RequestQueryMetaclass


class RequestQuery(NamedQuery, metaclass=RequestQueryMetaclass):
    """A :class:`cbra.query.NamedQuery` implementation that discovers its
    parameters from an incoming HTTP request.
    """
    __module__: str = 'cbra.query'

    @classmethod
    def get_query_parameters(cls) -> typing.List[Parameter]:
        params = []
        fields = cls.model.__fields__['args'].type_.__fields__
        for name, field in dict.items(fields):
            kwargs = {
                k: getattr(field.field_info, k)
                for k in field.field_info.__slots__
            }
            Source = fastapi.Path\
                if isinstance(field.field_info, PathField)\
                else fastapi.Query

            kwargs.setdefault('alias', name)
            params.append(
                Parameter(
                    name=name,
                    kind=Parameter.POSITIONAL_OR_KEYWORD,
                    default=Source(**kwargs),
                    annotation=field.type_
                )
            )
        return params

    def build(self,
        request: fastapi.Request,
        offset: int = fastapi.Query(default=0),
        limit: int = fastapi.Query(default=100),
        **kwargs
    ):
        """Build the query from the request parameters."""
        return self.model(
            limit=limit,
            offset=offset,
            args=kwargs
        )

    def get_signature(self) -> Signature:
        """Return a :class:`inspect.Signature` instance representing the
        call sigature of the dependency resolver.
        """
        sig = super().get_signature()
        params = list(sig.parameters.values())
        params.extend(self.get_query_parameters())
        return inspect.Signature([
            x for x in params
            if x.kind not in (Parameter.VAR_POSITIONAL, Parameter.VAR_KEYWORD)
        ])

    def get_request_dependencies(
        self
    ) -> typing.List[typing.Union[fastapi.params.Path, fastapi.params.Query]]:
        """Return the list of dependencies that are resolved from a
        :class:`fastapi.Request` instance.
        """
        return [
            Parameterless(x.name, x.default, x.annotation)
            for x in self.get_query_parameters()
        ]

    async def _run(self,
        runner: QueryRunner,
        request: fastapi.Request,
        offset: int,
        limit: int,
        *args,
        **kwargs
    ):
        return await super()._run(runner, request, offset, limit, **kwargs)



"""Declares :class:`OptionsMixin`."""
from typing import Any
from typing import Callable

import fastapi

import cbra
from cbra.types import ICorsPolicy
from cbra.types import IEndpoint
from cbra.cors import EndpointCorsPolicy
from .exceptions import Error
from .exceptions import LoginRequired


class Endpoint(cbra.Endpoint):
    cors_policy: type[ICorsPolicy] = EndpointCorsPolicy
    disable_options: bool = False
    error_url: str | None = None
    methods: set[str]
    name: str
    state: str | None = None
    summary: str
    options_description: str

    @classmethod
    def add_to_router(
        cls,
        *,
        app: IEndpoint.RouterType,
        base_path: str = '/',
        method: str | None = None,
        cors_policy: type[ICorsPolicy] | None = None,
        request_handler: Callable[..., IEndpoint.ResponseTypes] | None = None,
        **kwargs: Any
    ) -> None:
        name = kwargs.pop('name', None) or cls.name
        cors_policy = cls.cors_policy.configure(
            allowed_methods=cls.methods | {"OPTIONS"},
            allowed_headers={
                "Content-Type",
                "Wants-Digest"
            }
        )
        for method in cls.methods:
            super().add_to_router(
                app=app,
                base_path=base_path,
                cors_policy=cors_policy,
                method=method,
                request_handler=request_handler,
                name=name,
                **kwargs
            )

        if not cls.disable_options:
            Options = cbra.Endpoint.new(
                handle=cls.options
            )
            Options.add_to_router(
                app=app,
                base_path=base_path,
                cors_policy=cors_policy,
                request_handler=cls.options,
                method="OPTIONS",
                name=f'{name}.options',
                summary=cls.summary,
                description=cls.options_description,
                response_description="Allowed methods and CORS policy.",
                **kwargs
            )

    async def options(self) -> fastapi.Response:
        return fastapi.Response(
            headers={
                'Allow': str.join(',', sorted(self.cors_policy.allowed_methods))
             },
            status_code=200
        )

    @cbra.Endpoint.on_exception.register # type: ignore
    async def on_auth2_exception(
        self,
        exception: Error
    ) -> fastapi.Response:
        return await exception.as_response(
            request=self.request,
            state=self.state,
            error_url=self.error_url
        )

    @cbra.Endpoint.on_exception.register # type: ignore
    async def on_login_required(
        self,
        exception: LoginRequired
    ) -> fastapi.Response:
        return exception.as_response()

    @classmethod
    def exception_handler(cls, exception_class: type[Exception]):
        def decorator_factory(func: Callable[..., Any]):
            cbra.Endpoint.on_exception.register # type: ignore
            return func
        return decorator_factory


"""Declares :class:`RequestObjectDecoder`."""
from typing import cast
from typing import Any

import fastapi
from ckms.core.models import JSONWebSignature
from ckms.jose import PayloadCodec
from ckms.types import Malformed
from unimatrix.exceptions import CanonicalException

from .exceptions import Error
from .exceptions import NoSuchClient
from .params import ClientRepository
from .params import LocalIssuer
from .params import ServerCodec
from .types import IClientRepository


class RequestObjectDecoder:
    """Decores a **Request Object** as defined in :rfc:`9101`."""
    __module__: str = 'cbra.ext.oauth2'
    clients: IClientRepository
    codec: PayloadCodec
    issuer: str
    token_type: str = "oauth-authz-req+jwt"
    JARNotSigned: type[Error] = type('JARNotSigned', (Error,), {
        'error': 'invalid_request_object',
        'error_description': (
            "A JWT-Secured Authorization Request (JAR) must be "
            "signed using a key that is associated to the client."
        )
    })

    TokenTypeError: type[Error] = type('TokenTypeError', (Error,), {
        'error': 'invalid_request_object',
        'error_description': (
            "The JWT-Secured Authorization Request (JAR) was parsed "
            "correctly as a JSON Web Signature (JWS) but the \"typ\" "
            "header claim specified an invalid type. A JWS that "
            "represents a JAR must specify the \"oauth-authz-req+jwt\" "
            "in its header \"typ\" claim."
        )
    })

    InvalidAudience: type[Error] = type('InvalidAudience', (Error,), {
        'error': 'invalid_request_object',
        'error_description': (
            "The audience specified by the JWT-Secured Authorization "
            "Request (JAR) is not accepted by the server."
        )
    })

    InvalidIssuer: type[Error] = type('InvalidIssuer', (Error,), {
        'error': 'invalid_request_object',
        'error_description': (
            "The \"iss\" claim of the JWT-Secured Authorization Request (JAR) "
            "must match the \"client_id\" claim."
        )
    })

    InvalidRequest: type[Error] = type('InvalidRequest', (Error,), {
        'error': 'invalid_request_object',
        'error_description': (
            "The JWT-Secured Authorization Request (JAR) was succesfully "
            "parsed and the signature was valid for the specified client, "
            "but the claims indicate that the token was either not effective, "
            "expired or required claims were missing."
        )
    })

    MalformedRequest: type[Error] = type('MalformedRequest', (Error,), {
        'error': 'invalid_request_object',
        'error_description': (
            "The JWT-Secured Authorization Request (JAR) is malformed and "
            "could not be interpreted by the authorization server."
        )
    })

    InvalidSignature: type[Error] = type('InvalidSignature', (Error,), {
        'error': 'invalid_request_object',
        'error_description': "Signature validation failed."
    })

    def __init__(
        self,
        clients: IClientRepository = ClientRepository,
        codec: PayloadCodec = ServerCodec,
        issuer: str = LocalIssuer
    ) -> None:
        self.clients = clients
        self.codec = codec
        self.issuer = issuer

    async def decode_request(
        self,
        request: fastapi.Request
    ) -> dict[str, Any]:
        """Parses the incoming bytestream as a JWT-Secured Authorization Request
        (JAR).
        """
        return await self.decode_request_object(await request.body())

    async def decode_request_object(
        self,
        obj: bytes | str
    ) -> dict[str, Any]:
        if isinstance(obj, str):
            obj = str.encode(obj, 'ascii')
        try:
            jws, claims = await self.codec.jwt(obj, accept=self.token_type)
        except (Malformed, TypeError, ValueError):
            raise self.MalformedRequest
        if jws.typ != self.token_type:
            raise self.TokenTypeError
        if not bool(jws.signatures): # type: ignore
            raise self.JARNotSigned
        client_id = claims.extra.get('client_id')
        if claims.iss != client_id:
            raise self.InvalidIssuer
        if claims.iss is None or\
        not await self.clients.exists(client_id=claims.iss):
            raise NoSuchClient
        client = await self.clients.get(claims.iss)
        if not await client.verify_jws(jws):
            raise self.InvalidSignature
        try:
            claims.verify(
                audience={self.issuer},
                strict=True
            )
        except CanonicalException as exception:
            if exception.code == "WRONG_AUDIENCE":
                raise self.InvalidAudience
            raise self.InvalidRequest
        return claims.dict()


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import aorta
import fastapi
import pytest
from httpx import AsyncClient


class FooCommand(aorta.Command):
    foo: str


class BarCommand(aorta.Command):
    bar: str


class BazCommand(aorta.Command):
    baz: str


class TazCommand(aorta.Command):
    taz: str


class ErrorCommand(aorta.Command):
    foo: str


class FooBarCommandHandler(aorta.CommandHandler):

    async def handle( # type: ignore
        self,
        command: FooCommand | BarCommand
    ) -> None:
        pass


class TazCommandHander(aorta.CommandHandler):

    async def handle(
        self,
        command: TazCommand
    ) -> None:
        raise Exception


class ErrorCommandHandler(aorta.CommandHandler):

    def __init__(self, foo: str = fastapi.Path(...)):
        self.path = foo

    async def handle(self, command: ErrorCommand):
        pass


aorta.register(FooBarCommandHandler)
aorta.register(TazCommandHander)
aorta.register(ErrorCommandHandler)

@pytest.mark.asyncio
async def test_authenticate_with_token(
    id_token: str,
    client: AsyncClient
):
    command = FooCommand(foo='Hello world!')
    response = await client.post( # type: ignore
        url='/aorta',
        json={
            'subscription': 'projects/foobar/subscriptions/baztaz',
            'message': {
                'messageId': "id",
                'publishTime': "2014-10-02T15:01:23.045123456Z",
                'data': command.serialize()
            }
        },
        headers={'Authorization': f'Bearer {id_token}'}
    )
    assert response.status_code == 200
    assert not(response.text)


@pytest.mark.asyncio
async def test_failing_handler_returns_ok(
    id_token: str,
    client: AsyncClient
):
    command = TazCommand(taz='Hello world!')
    response = await client.post( # type: ignore
        url='/aorta',
        json={
            'subscription': 'projects/foobar/subscriptions/baztaz',
            'message': {
                'messageId': "id",
                'publishTime': "2014-10-02T15:01:23.045123456Z",
                'data': command.serialize()
            }
        },
        headers={'Authorization': f'Bearer {id_token}'}
    )
    assert response.status_code == 200, response.status_code
    assert not(response.text)
    assert 'X-Error-Code' in response.headers, response.headers


@pytest.mark.asyncio
async def test_unknown_object_type_sets_header(
    id_token: str,
    client: AsyncClient
):
    command = BazCommand(baz='Hello world!')
    response = await client.post( # type: ignore
        url='/aorta',
        json={
            'subscription': 'projects/foobar/subscriptions/baztaz',
            'message': {
                'messageId': "id",
                'publishTime': "2014-10-02T15:01:23.045123456Z",
                'data': command.serialize()
            }
        },
        headers={'Authorization': f'Bearer {id_token}'}
    )
    assert response.status_code == 200
    assert not(response.text)
    assert response.headers.get('Content-Length') in {'0', None}
    assert response.headers.get('X-Error-Code') == "UNKNOWN_MESSAGE_TYPE"


@pytest.mark.asyncio
async def test_malformed_object_type_sets_header(
    id_token: str,
    client: AsyncClient
):
    response = await client.post( # type: ignore
        url='/aorta',
        json={
            'subscription': 'projects/foobar/subscriptions/baztaz',
            'message': {
                'messageId': "id",
                'publishTime': "2014-10-02T15:01:23.045123456Z",
                'data': 'foo'
            }
        },
        headers={'Authorization': f'Bearer {id_token}'}
    )
    assert response.status_code == 200
    assert not(response.text)
    assert response.headers.get('Content-Length') in {'0', None}
    assert response.headers.get('X-Error-Code') == "MALFORMED_OBJECT"


@pytest.mark.asyncio
async def test_malformed_message_returns_no_body(
    id_token: str,
    client: AsyncClient
):
    response = await client.post( # type: ignore
        url='/aorta',
        json={
            'subscription': 'projects/foobar/subscriptions/baztaz',
            'message': {
                'messageId': "id",
                'publishTime': "2014-10-02T15:01:23.045123456Z",
            }
        },
        headers={'Authorization': f'Bearer {id_token}'}
    )
    assert response.status_code == 200
    assert not(response.text)
    assert response.headers.get('Content-Length') in {'0', None}


@pytest.mark.asyncio
async def test_improperly_configured_handler_is_suppressed(
    id_token: str,
    client: AsyncClient
):
    command = ErrorCommand(foo='Hello world!')
    response = await client.post( # type: ignore
        url='/aorta',
        json={
            'subscription': 'projects/foobar/subscriptions/baztaz',
            'message': {
                'messageId': "id",
                'publishTime': "2014-10-02T15:01:23.045123456Z",
                'data': command.serialize()
            }
        },
        headers={'Authorization': f'Bearer {id_token}'}
    )
    assert response.status_code == 200
    assert not(response.text)
    assert response.headers.get('Content-Length') in {'0', None}


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# pylint: skip-file
# type: ignore
from typing import Any

import pytest
from httpx import AsyncClient

from ..webhook.imessagetest import IMessageTest


class BaseMessageTest(IMessageTest):
    signature_header: str = 'x-picqer-signature'

    @pytest.mark.asyncio
    async def test_accept_event(
        self,
        client: AsyncClient
    ) -> None:
        params = self.get_request_params()
        response = await client.post(url='/', **params) # type: ignore
        dto = response.json()
        assert response.status_code == 200, response.text


"""Declares :class:`NullQueryModel`."""
from pydantic import dataclasses

from .basequerymodel import BaseQueryModel


@dataclasses.dataclass
class NullQueryModel(BaseQueryModel):
    """The default query model for endpoints that have no query parameters."""


import typing

import cbra
import cbra.cors


class ExampleEndpoint(cbra.Endpoint):

    async def handle(self) -> typing.Optional[typing.Dict[str, typing.Any]]:
        return {}


class ExampleCorsPolicy(cbra.cors.BaseCorsPolicy):
    allow_credentials: bool = True
    allowed_methods: typing.Set[str] = {"GET", "POST"}
    allowed_headers: typing.Set[str] = {"Authorization", "X-Custom-Header"}


def get_asgi_application() -> cbra.Application:
    app = cbra.Application()
    app.add(
        cbra.Options.new(
            allowed_methods={"GET", "POST", "OPTIONS"},
            cors_policy=ExampleCorsPolicy
        ),
        base_path='/'
    )
    app.add(
        ExampleEndpoint.new(cors_policy=ExampleCorsPolicy),
        base_path='/',
        method='GET'
    )
    app.add(
        ExampleEndpoint.new(cors_policy=ExampleCorsPolicy),
        base_path='/',
        method='POST'
    )
    return app


app = get_asgi_application()
if __name__ == '__main__':
    cbra.run('__main__:app', reload=True)


"""Declares :class:`CorsMixin`."""
import httpx
import pytest


class CorsMixin:
    cors_url: str
    cors_allowed_methods: set[str]
    cors_allowed_origin: str

    def get_cors_url(self) -> str:
        return self.cors_url

    @pytest.mark.asyncio
    async def test_preflight(self, client: httpx.AsyncClient, origin: str | None = None):
        origin = origin or self.cors_allowed_origin
        headers = {
            'Origin': origin,
            'Access-Control-Request-Methods': str.join(',', self.cors_allowed_methods)
        }
        response = await client.options( # type: ignore
            url=self.get_cors_url(),
            headers=headers
        )
        allowed_methods = ','.join(sorted(self.cors_allowed_methods | {"OPTIONS"}))
        assert response.status_code == 200, response.status_code
        assert 'Access-Control-Allow-Origin' in response.headers
        assert 'Access-Control-Allow-Methods' in response.headers
        assert response.headers['Access-Control-Allow-Origin'] == origin, response.headers
        assert response.headers['Access-Control-Allow-Methods'] == allowed_methods, response.headers

    @pytest.mark.asyncio
    async def test_preflight_accepts_none_whitelisted_domain(
        self,
        client: httpx.AsyncClient
    ):
        await self.test_preflight(client, "https://foo.invalid")


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest
import pytest_asyncio
from ckms.core import Keychain
from ckms.jose import PayloadCodec

from cbra import Application
from examples import rek


@pytest_asyncio.fixture(scope='session') # type: ignore
async def app() -> Application:
    app = rek.get_asgi_application()
    await app.on_startup()
    return app


@pytest.fixture
def keychain() -> Keychain:
    return rek.keychain


@pytest.fixture
def codec(keychain: Keychain) -> PayloadCodec:
    verifier = keychain.tagged('reks-signer')
    return PayloadCodec(verifier=verifier)

@pytest.fixture
def reks_url() -> str:
    return '/.well-known/request-encryption-key'


# pylint: skip-file
import httpx
import pytest

from ...servertest import ServerTest
from .iauthorizationcodeflow import IAuthorizationCodeFlow


class TestAuthorizationCodeFlow(IAuthorizationCodeFlow, ServerTest):
    client_id: str = 'test-authorization-code-flow'
    cors_url: str = ServerTest.authorize_url
    redirects: bool = True
    response_type: str = "code"

    @pytest.mark.asyncio
    async def test_request_and_request_uri_are_mutually_exclusive(
        self,
        client: httpx.AsyncClient
    ):
        await self.create_authorize_response(
            client=client,
            params={
                'client_id': self.client_id,
                'request_uri': 'foo',
                'request': 'bar'
            },
            status_code=400,
            error="invalid_request"
        )


class TestAuthorizationCodeFlowCORS(TestAuthorizationCodeFlow):
    origin = "https://allowed-cors-domain.example"
    redirects: bool = True


"""Declares :class:`BaseEnvelope`."""
import typing

import fastapi
import pydantic


class BaseEnvelope(pydantic.BaseModel):
    __module__: str = 'cbra.models'

    @classmethod
    def fromresult(cls,
        request: fastapi.Request,
        result: typing.Union[dict, list],
        params: dict
    ):
        """Instantiate a new envelope from the return value of a request
        handler.
        """
        if not isinstance(result, (dict, list)): # pragma: no cover
            raise TypeError("Return value must be a dictionary or list.")
        if isinstance(result, dict):
            params['spec'] = result
        if isinstance(result, list):
            params['items'] = [cls._envelope_factory(x) for x in result]
        return cls(**params)

    def get_path_parameter(self, request: fastapi.Request) -> typing.Union[int, str]:
        """Return the path parameter from this object."""
        return getattr(self.spec, self.Config.path_parameter)

    def is_list(self) -> bool:
        """Return a boolean indicating if the envelope contains a list."""
        return self._is_list

    def is_resource(self) -> bool:
        """Return a boolean indicating if the envelope contains a resource."""
        return self._is_resource

    def set_link(self, name: str, url: str):
        """Update the links attribute of the metadata."""
        self.metadata.links[name] = url



# pylint: skip-file
# type: ignore
import datetime
import secrets
import typing

import pytest
import pytest_asyncio
from ckms.jose import PayloadCodec
from ckms.types import JSONWebToken
from ckms.types import JSONWebKey
from httpx import AsyncClient

from cbra.ext.oauth2.types import GrantType
from cbra.ext.oauth2.types import ISubject
from cbra.utils import current_timestamp
from .conftest import get_subject_codec

class TokenMixin:
    base_url: str = "https://cbra-oauth2.localhost"
    default_scope: typing.Set[str] = {"read"}

    @pytest.fixture
    def token_audience(self) -> str:
        return f'{self.base_url}/oauth2/token'

    async def get_jwt_grant(self, subject: ISubject, **claims: typing.Any) -> str:
        codec = await get_subject_codec(subject)
        return await self.sign_jwt(codec, **claims)

    @pytest_asyncio.fixture # type: ignore
    async def grant(self, current_subject: ISubject, token_audience: str):
        return await self.get_jwt_grant(
            subject=current_subject,
            iss=current_subject.sub,
            aud=token_audience,
            sub=current_subject.sub
        )

    @pytest.fixture
    def scope(self) -> typing.Set[str]:
        return self.default_scope

    @pytest_asyncio.fixture # type: ignore
    async def token(
        self,
        client: AsyncClient,
        grant: str,
        current_subject: ISubject,
        scope: typing.Set[str]
    ):
        current_subject.client_scope['jwt'] = scope
        request = dict(
            client_id='jwt',
            grant_type=GrantType.jwt_bearer,
            assertion=grant
        )
        if scope:
            request['scope'] = str.join(' ', sorted(scope))
        response = await client.post(url='/oauth2/token', json=request)
        assert response.status_code == 200, response.text
        dto = response.json()
        return dto['access_token']

    async def sign_jwt(
        self,
        codec: PayloadCodec,
        **claims: typing.Any
    ) -> str:
        now = current_timestamp()
        defaults = {
            'jti': secrets.token_urlsafe(24),
            'iat': now,
            'exp': now+600,
            'nbf': now
        }
        defaults.update(claims)
        return await codec.encode(
            JSONWebToken(**defaults),
            signers=['sig']
        )


"""Declares :class:`AuthorizationCode`."""
import datetime

import pydantic


class AuthorizationCode(pydantic.BaseModel):
    __module__: str = 'cbra.ext.oauth2.types'
    created: datetime.datetime = pydantic.Field(
        default_factory=datetime.datetime.utcnow
    )

    code: str

    request_id: str


# pylint: skip-file
import pytest

from .. import __init__ as ioc
from ..models import DependencyList


@pytest.mark.asyncio
async def test_load(provider, package_name, symbol):
    dependencies = DependencyList(
        items=[
            {
                'spec': {
                    'type': 'file',
                    'name': 'ExampleDependency',
                    'path': 'VERSION'
                }
            },
            {
                'spec': {
                    'type': 'symbol',
                    'name': "SymbolDependency",
                    'qualname': f"{package_name}.conftest.SYMBOL"
                }
            },
        ]
    )

    await provider.load_many(dependencies)
    assert provider.is_satisfied('ExampleDependency')
    assert provider.is_satisfied('SymbolDependency')
    assert provider.resolve('SymbolDependency') == symbol



"""Declares :class:`StaticSubjectRepository`."""
from .settingssubjectrepository import SettingsSubjectRepository
from .types import ISubject


class StaticSubjectRepository(SettingsSubjectRepository):
    """A :class:`cbra.ext.oauth2.types.ISubjectRepository` implemenation
    that is configured with a static set of subjects.
    """
    __module__: str = 'cbra.ext.oauth2'
    subjects: dict[int | str, ISubject] = {}

    def __init__(self) -> None:
        self.subjects = type(self).subjects


"""Declares :class:`ResourceMetaclass`."""
import typing
from cbra.resource.iresource import IResource

import inflect
from pydantic import dataclasses

from cbra.endpointmetaclass import EndpointMetaclass
from .basefiltermodel import BaseFilterModel
from .pathparameter import PathParameter


engine = inflect.engine()
VOWEL_CHARS = {'a', 'e', 'i', 'o', 'u'}


class ResourceMetaclass(EndpointMetaclass):

    @staticmethod
    def get_article(word: str):
        return 'a' if str.lower(word[0]) not in VOWEL_CHARS else 'an'

    def __new__(
        cls,
        name: str,
        bases: typing.Tuple[typing.Type[type], ...],
        namespace: typing.Dict[str, typing.Any]
    ):
        super_new = super().__new__
        if namespace.get('__abstract__', False):
            return super_new(cls, name, bases, namespace)

        # Collect the attributes from the bases.
        base_attrs: dict[str, typing.Any] = {}
        for b in reversed(bases):
            if not isinstance(b, ResourceMetaclass):
                continue
            base_attrs.update(b.__dict__) # type: ignore

        # Construct the resource name from the name attribute, parent name or
        # class name.
        resource_name = namespace.get('name')\
            or base_attrs.get('name')\
            or str.replace(name, 'Resource', '')

        namespace.setdefault('name', name)
        namespace.setdefault('name_article', base_attrs.get('name_article') or cls.get_article(resource_name))
        namespace.setdefault('pluralname', engine.plural(resource_name))
        if 'verbose_name' not in namespace:
            namespace['verbose_name'] = base_attrs.get('verbose_name') or str.title(namespace['name'])
        if 'verbose_name_plural' not in namespace:
            namespace['verbose_name_plural'] = engine.plural(str.title(namespace['verbose_name']))
        # If resource class is a subresource, then a parent member is
        # contained by the namespace.
        parent: ResourceMetaclass = namespace.get(
            'parent',
            base_attrs.get('parent')
        )

        # Construct the path parameters, including those defined by
        # the parent classes.
        if 'path_parameter' in namespace:
            path_parameter = namespace['path_parameter'] = PathParameter(
                name=namespace['path_parameter'],
                annotation=namespace.pop('path_parameter_class', None)
            )
        else:
            path_parameter = base_attrs.get('path_parameter')

        # Construct a pydantic model holding the query parameters.
        annotations = namespace.setdefault('__annotations__', {})
        #if 'query_parameters' in namespace and not namespace.get('query_model'):
        #    model_attrs: typing.Dict[typing.Any, typing.Any] = {}
        #    model_hints: typing.Dict[typing.Any, typing.Any] = {}
        #    for i, param in enumerate(namespace.pop('query_parameters', [])):
        #        if param.alias is None:
        #            raise ValueError(
        #                f"{name}.query_parameters[{i}] must set the "
        #                "'alias' parameter."
        #            )
        #        if QueryOptions.is_reserved_field(param.alias):
        #            raise ValueError(f"The name '{param.alias}' is reserved.")
        #        if not isinstance(param, fastapi.params.Query):
        #            raise NotImplementedError
        #        model_attrs[param.alias] = param
        #        model_hints[param.alias] = param.extra.pop('annotation', None) or str
        #    if model_attrs:
        #        annotations['query'] = namespace['query'] = dataclasses.dataclass(
        #            type(
        #                f'EndpointQuery',
        #                (BaseQueryModel,),
        #                {**model_attrs, '__annotations__': model_hints}
        #            )
        #        )

        attrs: dict[str, typing.Any] = {}
        for base in reversed(bases):
            attrs.update(base.__dict__) # type: ignore
        attrs.update(namespace)

        # Construct the filter model, which is used by the actions listed
        # by the filter_actions attribute.
        filter_name = str.replace(name, 'Resource', '')
        filter_model = type(f'{filter_name}FilterModel', (BaseFilterModel,), {})
        if attrs.get('searchable'):
            filter_params = namespace.pop('filter_defaults', [])
            for base in bases:
                if not hasattr(base, 'filter_defaults'):
                    continue
                filter_params.extend(base.filter_defaults) # type: ignore
            filter_attrs = {}
            for filter_param in filter_params:
                filter_param.add_to_namespace(filter_attrs)
            filter_model = type(f'{filter_name}FilterModel', (BaseFilterModel,), filter_attrs)
        annotations['query'] = namespace['query'] = dataclasses.dataclass(filter_model)

        # Create the new class with the updated namespace.
        new_class: type[IResource] = typing.cast(
            type[IResource],
            super_new(cls, name, bases, {
                **namespace,
            })
        )

        # The path parameter is required to determine the path for the
        # collection endpoints.
        if new_class.subresources and not new_class.path_parameter:
            raise ValueError(
                f"{name}.path_parameter can not be None because "
                f"{name}.subresources is a non-empty list."
            )

        # Set the parent attribute on the subresources. This is used to
        # determine the path parameters.
        for i, child in enumerate(new_class.subresources):
            new_class.subresources[i] = typing.cast(
                type[IResource],
                child.new(parent=new_class)
            )

        return new_class


# pylint: skip-file
import httpx
import pytest

import cbra
from cbra.const import REQUEST_METHODS
from examples.cors import get_asgi_application
from examples.cors import ExampleCorsPolicy


ALLOWED_HEADERS = ExampleCorsPolicy.allowed_headers
ALLOWED_METHODS = REQUEST_METHODS & ExampleCorsPolicy.allowed_methods
FORBIDDEN_HEADERS = {'X-Foo', 'X-Bar'}
FORBIDDEN_METHODS = REQUEST_METHODS - ALLOWED_METHODS
ORIGIN = "https://unimatrixone.local"


@pytest.fixture
def app() -> cbra.Application:
    ExampleCorsPolicy.allowed_origins.add(ORIGIN)
    return get_asgi_application()


@pytest.mark.parametrize("method", sorted(ALLOWED_METHODS))
@pytest.mark.asyncio
async def test_allowed_methods_are_accepted_by_options(client: httpx.AsyncClient, method: str):
    response = await client.options(
        url='/',
        headers={
            'Origin': ORIGIN,
            'Access-Control-Request-Methods': method
        }
    )
    assert response.status_code == 200
    assert 'Access-Control-Allow-Credentials' in response.headers
    assert 'Access-Control-Allow-Headers' in response.headers
    assert 'Access-Control-Allow-Methods' in response.headers
    assert 'Access-Control-Allow-Origin' in response.headers
    assert response.headers['Access-Control-Allow-Origin'] == ORIGIN


@pytest.mark.parametrize("header", sorted(ALLOWED_HEADERS))
@pytest.mark.asyncio
async def test_allowed_methods_are_accepted_by_options(client: httpx.AsyncClient, header: str):
    response = await client.options(
        url='/',
        headers={
            'Origin': ORIGIN,
            'Access-Control-Request-Headers': header
        }
    )
    assert response.status_code == 200
    assert 'Access-Control-Allow-Credentials' in response.headers
    assert 'Access-Control-Allow-Headers' in response.headers
    assert 'Access-Control-Allow-Methods' in response.headers
    assert 'Access-Control-Allow-Origin' in response.headers
    assert response.headers['Access-Control-Allow-Origin'] == ORIGIN
    assert header in response.headers['Access-Control-Allow-Headers']


@pytest.mark.parametrize("method", sorted(ALLOWED_METHODS))
@pytest.mark.asyncio
async def test_allowed_methods_are_accepted(client: httpx.AsyncClient, method: str):
    response = await getattr(client, method.lower())(
        url='/',
        headers={'Origin': ORIGIN}
    )
    assert response.status_code == 200
    assert 'Access-Control-Allow-Credentials' in response.headers, response.headers
    assert 'Access-Control-Expose-Headers' in response.headers
    assert 'Access-Control-Allow-Origin' in response.headers
    assert response.headers['Access-Control-Allow-Origin'] == ORIGIN


@pytest.mark.parametrize("header", sorted(ALLOWED_HEADERS))
@pytest.mark.asyncio
async def test_allowed_headers_are_accepted(client: httpx.AsyncClient, header: str):
    response = await client.get(
        url='/',
        headers={'Origin': ORIGIN, header: 'Foo'}
    )
    assert response.status_code == 200
    assert 'Access-Control-Allow-Credentials' in response.headers, response.headers
    assert 'Access-Control-Expose-Headers' in response.headers
    assert 'Access-Control-Allow-Origin' in response.headers
    assert response.headers['Access-Control-Allow-Origin'] == ORIGIN
    assert header in response.headers['Access-Control-Expose-Headers']


@pytest.mark.parametrize("method", sorted(FORBIDDEN_METHODS))
@pytest.mark.asyncio
async def test_forbidden_methods_are_rejected_by_options(client: httpx.AsyncClient, method: str):
    response = await client.options(
        url='/',
        headers={
            'Origin': ORIGIN,
            'Access-Control-Request-Methods': method
        }
    )
    assert response.status_code == 200
    assert 'Access-Control-Allow-Credentials' in response.headers
    assert 'Access-Control-Allow-Headers' in response.headers
    assert 'Access-Control-Allow-Methods' in response.headers
    assert 'Access-Control-Allow-Origin' in response.headers
    assert 'Access-Control-Max-Age' in response.headers
    assert method not in response.headers['Access-Control-Allow-Methods']


@pytest.mark.parametrize("header", sorted(FORBIDDEN_HEADERS))
@pytest.mark.asyncio
async def test_forbidden_headers_are_rejected_by_options(client: httpx.AsyncClient, header: str):
    response = await client.options(
        url='/',
        headers={
            'Origin': ORIGIN,
            'Access-Control-Request-Headers': header
        }
    )
    assert response.status_code == 200
    assert 'Access-Control-Allow-Credentials' in response.headers
    assert 'Access-Control-Allow-Headers' in response.headers
    assert 'Access-Control-Allow-Methods' in response.headers
    assert 'Access-Control-Allow-Origin' in response.headers
    assert 'Access-Control-Max-Age' in response.headers
    assert header not in response.headers['Access-Control-Allow-Headers']


@pytest.mark.asyncio
async def test_forbidden_origin_is_rejected_by_options(client: httpx.AsyncClient):
    response = await client.options(
        url='/',
        headers={
            'Origin': "https://example.com",
        }
    )
    assert response.status_code == 200
    assert 'Access-Control-Allow-Credentials' not in response.headers
    assert 'Access-Control-Allow-Headers' not in response.headers
    assert 'Access-Control-Allow-Methods' not in response.headers
    assert 'Access-Control-Allow-Origin' not in response.headers
    assert 'Access-Control-Max-Age' not in response.headers


@pytest.mark.parametrize("method", ALLOWED_METHODS)
@pytest.mark.asyncio
async def test_no_origin_allows_all_methods(client: httpx.AsyncClient, method: str):
    response = await getattr(client, method.lower())(
        url='/'
    )
    assert response.status_code == 200
    assert 'Access-Control-Allow-Credentials' not in response.headers
    assert 'Access-Control-Allow-Headers' not in response.headers
    assert 'Access-Control-Allow-Methods' not in response.headers
    assert 'Access-Control-Allow-Origin' not in response.headers
    assert 'Access-Control-Max-Age' not in response.headers


@pytest.mark.parametrize("method", ALLOWED_METHODS)
@pytest.mark.asyncio
async def test_same_origin_allows_all_methods(client: httpx.AsyncClient, method: str):
    response = await getattr(client, method.lower())(
        url='/',
        headers={'Origin': str(client.base_url)}
    )
    assert response.status_code == 200
    assert 'Access-Control-Allow-Credentials' not in response.headers
    assert 'Access-Control-Allow-Headers' not in response.headers
    assert 'Access-Control-Allow-Methods' not in response.headers
    assert 'Access-Control-Allow-Origin' not in response.headers
    assert 'Access-Control-Max-Age' not in response.headers


"""Declares :class:`IClient`."""
from typing import cast
from typing import Any
from typing import Union

import ckms
import ckms.jose
import pydantic
from ckms.core.models import JSONWebSignature
from ckms.types import IKeychain
from ckms.types import JSONWebToken

from .granttype import GrantType
from .isubject import ISubject
from .redirecturl import RedirectURL
from .responsetype import ResponseType


class IClient(pydantic.BaseModel):
    """Specifies the interface of OAuth 2.0 client objects.

    An OAuth 2.0 client can be determined from various parameters in an
    HTTP request.

    For requests to the :term:`Authorization Endpoint`, the client is
    specified by the ``client_id`` parameter. This is the case for both
    public clients and confidential clients.

    The client is determined for :term:`Token Endpoint` requests by
    either the ``client_id`` parameter in the request body if the
    client is public, or by the credentials provided in the
    ``Authorization`` header. In the latter case, the ``client_id``
    parameter is absent in the request body.
    """
    __module__: str = 'cbra.ext.oauth2.types'
    __abstract__: bool = True

    client_id: str

    @classmethod
    async def fromquery(
        cls,
        clients: type,
        required: bool = False
    ) -> Any:
        """Return a callable that resolves the current client from the
        query parameters.
        """
        raise NotImplementedError

    @classmethod
    def fromdict(cls, params: dict[str, Any]) -> 'IClient':
        """Construct a new client from a dictionary."""
        raise NotImplementedError

    @classmethod
    def configure(cls, **kwargs: Any) -> 'IClient':
        """Configures the class with the given parameters."""
        return cast(IClient, type(cls.__name__, (cls,), kwargs))

    def as_subject(self) -> ISubject:
        """Return a :class:`~cbra.ext.oauth2.types.ISubject` implementation
        representing the client itself. For use with the client credentials
        grant.
        """
        raise NotImplementedError

    def allows_audience(self, issuer: str, audience: set[str]) -> bool:
        """Return a boolean indicating if the client allows the issueing
        of access token for the given audiences set `audience`.
        """
        raise NotImplementedError

    def allows_scope(self, scope: set[str]) -> bool:
        """Return a boolean indicating if the client allows the given scope."""
        raise NotImplementedError

    def allows_grant_type(self, grant_type: GrantType) -> bool:
        """Return a boolean indicating if the client allows the given grant type."""
        raise NotImplementedError

    def allows_origin(self, origin: str | None) -> bool:
        """Return a boolean indicating if the client allows the given origin in the
        context of Cross-Origin Resource Sharing (CORS).
        """
        raise NotImplementedError

    def allows_response_type(self, response_type: ResponseType) -> bool:
        """Return a boolean indicating if the client allows the given response
        type.
        """
        raise NotImplementedError

    def can_issue_multiple(self) -> bool:
        """Return a boolean indicating if the client allows issueing
        access tokens to multiple audiences.
        """
        raise NotImplementedError

    def get_default_redirect(self) -> str:
        """Return the default redirect URL, if specified."""
        raise NotImplementedError

    def get_consent_url(self, scope: set[str]) -> str:
        """Return the URL at which the resource owner can grant consent
        for the given scope to this client.
        """
        raise NotImplementedError

    def get_redirect_url(
        self,
        url: Union[str, RedirectURL] | None = None,
        fatal: bool = True
    ) -> RedirectURL | None:
        raise NotImplementedError

    def is_confidential(self) -> bool:
        """Return a boolean indicating if the client is confidential and
        must authenticate itself.
        """
        raise NotImplementedError

    def is_first_party(self) -> bool:
        """Return a boolean indicating if this is a first-party client
        that does not have to obtain consent.
        """
        raise NotImplementedError

    def is_public(self) -> bool:
        """Return a boolean indicating if the client is public."""
        raise NotImplementedError

    def assigns_ppid(self) -> bool:
        """Return a boolean indicating if this client assigns a PPID."""
        raise NotImplementedError

    async def issue_token(
        self,
        *,
        algorithm: str,
        codec: ckms.jose.PayloadCodec,
        issuer: str,
        audience: list[str] | str | set[str] | None,
        subject: ISubject,
        using: str,
        ttl: int,
        scope: set[str]
    ) -> str:
        """Issue an access token encoded as a JSON Web Token (JWT)
        as defined in :rfc:`9068`. Return a string containing the
        token.
        """
        raise NotImplementedError

    async def verify_jws(
        self,
        jws: JSONWebSignature
    ) -> bool:
        """Verify a :class:`~JSONWebSignature` using the public
        keys associated to the client.
        """
        raise NotImplementedError

    async def verify_jwt(
        self,
        token: Union[bytes, str],
        decrypter: IKeychain | None,
        compact_only: bool = False,
        audience: str | None = None
    ) -> JSONWebToken:
        """Verifies a JSON Web Token (JWT) using the public key of the
        client.

        Args:
            token (byte-sequence or string): the JWT to deserialize and
                verify.
            decrypter: a :class:`~IKeychain` implementation that has the
                private keys to decrypt, if the payload is a JWE.
            compact_only (bool): indicates if only compact deserialization
                is allowed. If `compact_only` is ``True`` and the token
                is not compact serialized, an exception is raised.

        Returns
            A :class:`dict` containing the claims in the JWT.
        """
        raise NotImplementedError


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import TypeAlias

from .event import Event


__all__: list[str] = [
    'Event'
]

InvoicePaid: type[Event] = Event.new(
    name='InvoicePaid',
    event="sales_invoice_state_changed_to_paid"
)

InvoiceDefaulted: type[Event] = Event.new(
    name='InvoiceDefaulted',
    event="sales_invoice_state_changed_to_late"
)


EventType: TypeAlias = (
    InvoicePaid | InvoiceDefaulted
)


"""Declares :class:`OpenAuthorizationEndpoint`."""
import inspect
import typing

import cbra
from .iopenauthorizationserver import IOpenAuthorizationServer


class OpenAuthorizationEndpoint(cbra.handler.BaseRequestHandler):
    __module__: str = 'cbra.ext.oauth2'
    server: IOpenAuthorizationServer
    handler_class: typing.Type[typing.Any]
    handler: typing.Callable[..., typing.Any]

    @property
    def __signature__(self) -> inspect.Signature:
        return inspect.signature(self.handler)

    def __init__(
        self,
        server: IOpenAuthorizationServer,
        handler_class: typing.Type[typing.Any]
    ):
        self.server = server
        self.handler_class = handler_class
        self.handler = handler_class.as_handler(
            server=self.server,
            metadata=self.server.metadata
        )

    async def __call__(self, *args, **kwargs) -> typing.Any:
        """Resolve dependencies and invoke the handler."""
        return await self.handler(*args, **kwargs)


"""Declares :func:`TokenRequestClient`."""
import typing

import fastapi
from ckms.types import ClaimSet
from fastapi.security.utils import get_authorization_scheme_param
from unimatrix.exceptions import CanonicalException

from .exceptions import ClientAuthenticationRequired
from .exceptions import InvalidClient
from .exceptions import InvalidClientAuthenticationScheme
from .exceptions import InvalidGrant
from .exceptions import NoSuchClient
from .types import BaseGrant
from .types import ClientAssertionType
from .types import IClientRepository
from .types import IStorage
from .types import TokenRequestParameters
from .params import ClientRepository
from .params import TokenEndpointURL
from .params import TransientStorage


__all__ = ['TokenRequest']


async def get_token_request(
    request: TokenRequestParameters,
    authorization: typing.Optional[str] = fastapi.Header(
        default=None,
        alias='Authorization',
        description=(
            "The credentials of the client, if the client is confidential "
            "and the credentials are not provided using a client "
            "assertion or the `client_id` and `client_secret` request "
            "parameters."
        )
    ),
    clients: IClientRepository = ClientRepository,
    storage: IStorage = TransientStorage,
    token_endpoint: str = TokenEndpointURL
) -> BaseGrant:
    """Retrieve the current client from either the ``client_id`` specified
    in the request body or the ``Authorization`` header (for confidential
    clients).
    """
    dto = request.get_root()
    client_id = dto.client_id
    grant_type = dto.grant_type

    # Requests to the Token Endpoint always have a client specified. If either
    # the header or the client_id in the request body is missing, we are in an
    # error state.
    if not authorization and not client_id and not dto.has_client_assertion():
        raise InvalidClient(
            error="invalid_client",
            error_description=(
                "Specify the `client_id` parameter in the request body if "
                "the client is public, else provide credentials with the "
                "`Authorization` header."
            )
        )
    if (client_id is not None and not await clients.exists(client_id)):
        raise NoSuchClient

    # If the client attempted to authenticate via the "Authorization" 
    # request header field, the authorization server MUST respond with
    # an HTTP 401 (Unauthorized) status code and include the
    # "WWW-Authenticate" response header field matching the
    # authentication scheme used by the client (RFC 6749, Section 5.2).
    client = None
    scheme, credentials = get_authorization_scheme_param(authorization or '')
    if scheme and credentials:
        if str.lower(scheme) != "bearer":
            raise InvalidClientAuthenticationScheme(
                using=scheme,
                required="Bearer"
            )
        raise NotImplementedError
        # client = await clients.fromcredentials(credentials.credentials)

    # If the client_id variable is None at this point, then it was not provided
    # in the request body, could not be determined from the Authorization header,
    # and there was no client assertion or the client assertion did not verify.
    if client_id is None: # pragma: no cover
        raise NoSuchClient

    # A client MAY use the "client_id" request parameter to identify itself
    # when sending requests to the token endpoint.  In the "authorization_code"
    # "grant_type" request to the token endpoint, an unauthenticated client
    # MUST send its "client_id" to prevent itself from inadvertently accepting
    # a code intended for a client with a different "client_id".  This protects
    # the client from substitution of the authentication code
    # (RFC 6749, Section 3.2.1).
    client = await clients.get(client_id)

    # Authenticate the client using a client assertion, if one was provided.
    if client.is_confidential() and dto.has_client_assertion():
        if dto.client_assertion_type != ClientAssertionType.jwt_bearer:
            raise NotImplementedError
        assert dto.client_assertion is not None # nosec
        try:
            jwt = await client.verify_jwt(
                token=dto.client_assertion,
                decrypter=None,
                audience=token_endpoint
            )
        except (ValueError, TypeError):
            raise InvalidGrant(
                error_description=(
                    "The supplied credential was malformed and could not be "
                    "interpreted by the authorization server."
                )
            )
        except CanonicalException as exception:
            if exception.code in {"INVALID_SIGNATURE", "UNVERIFIABLE"}:
                raise InvalidGrant(
                    error_description=(
                        "The JSON Web Token (JWT) was signed using a key that is "
                        "not accepted by the client."
                    )
                )
            elif exception.code == "WRONG_AUDIENCE":
                raise InvalidGrant(
                    error_description=(
                        "The JSON Web Token (JWT) specified an audience that "
                        "is not accepted by the token endpoint. Ensure that "
                        f"the \"aud\" claim is equal to \"{token_endpoint}\"."
                    )
                )
            raise
        else:
            if jwt.sub != client_id:
                raise InvalidClient(
                    error="invalid_client",
                    error_description=(
                        "The \"sub\" claim in the assertion does not match "
                        "the \"client_id\" specified in the request."
                    )
                )
            elif jwt.iss != client_id:
                raise InvalidClient(
                    error="invalid_client",
                    error_description=(
                        "The \"iss\" claim in the assertion does not match "
                        "the \"client_id\" specified in the request."
                    )
                )
            if await storage.consume(jwt):
                raise InvalidClient(
                    error="invalid_client",
                    error_description=(
                        "A client assertion can only be used once."
                    )
                )

    elif not client.is_public():
        # Other methods are not supported.
        raise ClientAuthenticationRequired

    # Ensure that the client allows the requested grant.
    if not client.allows_grant_type(grant_type):
        raise InvalidGrant(
            error_description=(
                "The client does not allow the given grant: "
                f"{grant_type.value}."
            )
        )

    dto.set_client(client)
    return dto


TokenRequest = fastapi.Depends(get_token_request)


"""Declares :class:`IOpenAuthotizationServer`."""
import typing

import fastapi

from .iclient import IClient
from .iclientrepository import IClientRepository
from .servermetadata import ServerMetadata
from .tokentype import TokenType


class IOpenAuthorizationServer:
    """The interface of a class that implements an OAuth 2.0
    server.
    """
    __module__: str = 'cbra.ext.oauth2.types'
    subject_token_verifiers: typing.Dict[str, typing.Any]

    #: A class that is used as the repository for clients.
    clients: type[IClientRepository]

    async def deserialize_jwt(
        self,
        client: IClient,
        token: typing.Union[bytes, str]
    ) -> typing.Dict[str, typing.Any]:
        raise NotImplementedError

    async def get_subject_token_verifier(
        self,
        token_type: TokenType
    ) -> typing.Any:
        """Return a :class:`~cbra.ext.oauth2.BaseSubjectTokenVerifier` that
        is used to verify the ``subject_token`` parameter during an
        :rfc:`8693` OAuth 2.0 Token Exchange.
        """
        raise NotImplementedError

    async def retrieve_request(
        self,
        client: IClient,
        request_uri: str
    ) -> str:
        raise NotImplementedError

    def register_keys(
        self,
        default_signing_key: str,
        default_encryption_key: str,
        keys: typing.Dict[str, typing.Dict[str, typing.Any]]
    ) -> None:
        raise NotImplementedError


# pylint: skip-file
# type: ignore
import pytest
import pytest_asyncio
from ckms.core import Keychain
from ckms.jose import PayloadCodec
from ckms.types import JSONWebKey
from ckms.types import JSONWebToken
from cryptography.hazmat.primitives.asymmetric import ed448

from cbra.ext import oauth2


@pytest.fixture(scope='session')
def keychain() -> Keychain:
    return Keychain()


@pytest.fixture(scope='session')
def private_key():
    return ed448.Ed448PrivateKey.generate()


@pytest_asyncio.fixture(scope='session')
async def jwk(keychain, private_key):
    keychain.configure({
        'sig': {
            'provider': "local",
            'key': {'content': private_key},
            'alg': "EdDSA",
            'kty': "OKP",
            'use': 'sig',
            'crv': 'Ed448'
        }
    })
    await keychain
    spec = keychain.get('sig')
    return spec.as_jwk()


@pytest.fixture
def codec(keychain):
    return PayloadCodec(
        verifier=keychain,
        signer=keychain
    )


def test_allowed_redirect_urls_are_parsed_from_strings():
    config = oauth2.ClientConfig(
        client_id="foo",
        redirect_uris=["https://www.example.com"]
    )
    assert isinstance(config.redirect_uris, list)
    assert isinstance(config.redirect_uris[0], oauth2.types.RedirectURL)
    assert str(config.redirect_uris[0]) == "https://www.example.com"


def test_default_redirect_url_is_parsed_from_string():
    config = oauth2.ClientConfig(
        client_id="foo",
        default_redirect_url="https://www.example.com"
    )
    assert isinstance(config.default_redirect_url, oauth2.types.RedirectURL)


def test_can_redirect_accepts_only_whitelisted_urls():
    config = oauth2.ClientConfig(
        client_id="foo",
        redirect_uris=["https://www.example.com"]
    )
    assert config.can_redirect("https://www.example.com")
    assert not config.can_redirect("https://www.example.net")


def test_can_redirect_accepts_default():
    config = oauth2.ClientConfig(
        client_id="foo",
        default_redirect_url="https://www.example.com"
    )
    assert config.can_redirect("https://www.example.com")
    assert not config.can_redirect("https://www.example.net")


def test_get_redirect_url_with_whitelisted_url():
    config = oauth2.ClientConfig(
        client_id="foo",
        redirect_uris=["https://www.example.com"]
    )
    url = config.get_redirect_url("https://www.example.com")
    assert url == "https://www.example.com"


def test_get_redirect_url_with_non_whitelisted_url():
    config = oauth2.ClientConfig(
        client_id="foo",
        redirect_uris=["https://www.example.com"]
    )
    with pytest.raises(oauth2.exceptions.RedirectForbidden):
        url = config.get_redirect_url("https://www.example.net")


def test_default_redirect_is_added_to_allowed_redirect_urls():
    config = oauth2.ClientConfig(
        client_id="foo",
        default_redirect_url="https://www.example.com"
    )
    assert "https://www.example.com" in config.redirect_uris


def test_get_redirect_url_with_none_returns_default():
    config = oauth2.ClientConfig(
        client_id="foo",
        default_redirect_url="https://www.example.com"
    )
    url = config.get_redirect_url(None)


def test_get_redirect_url_with_whitelisted_str_returns_redirect():
    config = oauth2.ClientConfig(
        client_id="foo",
        redirect_uris=["https://www.example.net"],
        default_redirect_url="https://www.example.com"
    )
    url = config.get_redirect_url("https://www.example.net")
    assert url == "https://www.example.net"


def test_get_redirect_url_with_whitelisted_redirecturi_returns_redirect():
    config = oauth2.ClientConfig(
        client_id="foo",
        redirect_uris=["https://www.example.net"],
        default_redirect_url="https://www.example.com"
    )
    url = config.get_redirect_url(oauth2.types.RedirectURL.validate("https://www.example.net"))
    assert url == "https://www.example.net"


def test_get_redirect_url_with_none_and_no_default():
    config = oauth2.ClientConfig(
        client_id="foo",
        redirect_uris=["https://www.example.com", "https://www.example.localhost"]
    )
    with pytest.raises(oauth2.exceptions.MissingRedirectURL):
        config.get_redirect_url(None)


@pytest.mark.asyncio
async def test_verify_valid_jwt_with_client(jwk, codec):
    config = oauth2.ClientConfig(
        client_id="foo",
        keys=[jwk]
    )
    jws = await codec.encode(
        JSONWebToken(foo=1),
        signers=['sig']
    )
    assert await config.verify_jwt(jws.encode(), None)


@pytest.mark.asyncio
async def test_verify_invalid_jwt_with_client_raises(jwk, codec):
    config = oauth2.ClientConfig(
        client_id="foo",
        keys=[jwk]
    )
    jws = await codec.encode(
        JSONWebToken(foo=1),
        signers=['sig']
    )
    with pytest.raises(ValueError):
        assert not await config.verify_jwt(jws + 'd', None)




# pylint: skip-file
import pytest

from .provider import Provider


SYMBOL = "Hello world!"


@pytest.fixture
def symbol():
    return SYMBOL


@pytest.fixture
def package_name():
    return str.replace(__name__, '.conftest', '')


@pytest.fixture
def provider():
    return Provider()



"""Declares :class:`PreparedRequest`."""
import typing

import fastapi
import fastapi.params
from ckms.jose.exceptions import JOSEException

from cbra.ext.oauth2.iclient import IClient
from .authorizationrequest import AuthorizationRequest
from .codechallenge import CodeChallenge
from .exceptions import Error
from .iopenauthorizationserver import IOpenAuthorizationServer


class PreparedAuthorizationRequest(AuthorizationRequest):
    """A prepared request that is supplied by the client through the
    ``request`` or ``request_uri`` parameter.
    """
    __module__: str = 'cbra.ext.oauth2'

    request: typing.Optional[str] = None
    request_uri: typing.Optional[str] = None

    @classmethod
    def as_dependant(
        cls,
        server: IOpenAuthorizationServer,
        client: fastapi.params.Depends
    ) -> typing.Any:
        """Construct a function with the appropriate signature to inject the
        parameters and return a :class:`PreparedAuthorizationRequest`.
        """
        async def f(
            client: typing.Union[fastapi.params.Depends, IClient] = client,
            dto: AuthorizationRequest = AuthorizationRequest.as_dependant(
                client=client,
                pkce=CodeChallenge.as_dependant()
            ),
            request: str = fastapi.Query(
                default=None,
                title="Request",
                description=(
                    "A JSON Web Token (JWT) that holds an OAuth 2.0 authorization "
                    "request as JWT Claims Set."
                )
            ),
            request_uri: str = fastapi.Query(
                default=None,
                title="Request URI",
                description=(
                    "Absolute URI from which the **Request Object** can be obtained."
                )
            )
        ) -> typing.Any:
            if not request and not request_uri:
                return dto
            params = {
                **await cls.deserialize_par(
                    server=server,
                    client=client,
                    request=await cls.fetch_par(server, client, request_uri)
                ),
                **await cls.deserialize_par(server, client, request),
                **dto.dict(exclude_defaults=True)
            }
            return cls(**params)

        return fastapi.Depends(f)

    @staticmethod
    async def deserialize_par(
        server: IOpenAuthorizationServer,
        client: IClient,
        request: typing.Optional[str] = None
    ) -> typing.Dict[str, typing.Any]:
        """Deserialize the content of a Prepared Authorization Request (PAR)."""
        if request is None:
            return {}
        try:
            return await server.deserialize_jwt(client, request)
        except JOSEException:
            raise Error(
                error_description="The signature of the request could not be verified."
            )

    @staticmethod
    async def fetch_par(
        server: IOpenAuthorizationServer,
        client: IClient,
        request_uri: typing.Optional[str] = None
    ) -> typing.Optional[str]:
        """Fetch a Prepared Authorization Request (PAR) from the remote URI."""
        if request_uri is None:
            return None

        return await server.retrieve_request(client,  request_uri)

    def __bool__(self) -> bool:
        return bool(self.request or self.request_uri)


# pylint: skip-file
import itertools
import json
import typing
import urllib.parse

import httpx
import pydantic
import pytest
import pytest_asyncio
import yaml

import cbra
from . import BarModel
from . import FooModel
from . import ContentNegotiationEndpoint


SERIALIZE = {
    "application/json": json.dumps,
    "application/yaml": yaml.safe_dump,
    "application/x-www-form-urlencoded": urllib.parse.urlencode
}

SUPPORTED_TYPES = [
    "application/json",
    "application/yaml",
    "application/x-www-form-urlencoded"
]

VALID_REQUESTS = itertools.product(
    SUPPORTED_TYPES,
    [
        FooModel(foo=1),
        BarModel(bar=2)
    ]
)


@pytest.fixture
def app() -> cbra.Application:
    app = cbra.Application()
    app.add(ContentNegotiationEndpoint, base_path='/', method="POST")
    return app


@pytest_asyncio.fixture
async def client(
    app: cbra.Application
) -> typing.AsyncGenerator[httpx.AsyncClient, None]:
    params = {
        'app': app,
        'base_url': "https://cbra.localhost"
    }
    async with httpx.AsyncClient(**params) as c:
        yield c


@pytest.mark.parametrize("content_type,dto", VALID_REQUESTS)
@pytest.mark.asyncio
async def test_content_is_deserialized_and_casted(
    client: httpx.AsyncClient,
    content_type: str,
    dto: pydantic.BaseModel
):
    content: bytes = str.encode(SERIALIZE[content_type](dto.dict()))
    response = await client.post(
        url='/',
        content=content,
        headers={
            'Accept': 'application/json',
            'Content-Type': content_type
        }
    )
    assert response.status_code == 200
    assert response.json() == dto.dict()


# Copyright (C) 2022 Cochise Ruhulessin <cochiseruhulessin@gmail.com>
# 
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
"""Declares :class:`GoogleEndpoint`."""
import fastapi
import fastapi.params
import httpx
from ckms.jose import PayloadCodec
from ckms.types import JSONWebToken
from ckms.types import MalformedPayload
from ckms.types import ServerMetadata
from fastapi.security.utils import get_authorization_scheme_param

from cbra import Endpoint
from cbra.cors import NullCorsPolicy
from cbra.exceptions import AuthenticationRequired
from cbra.exceptions import Forbidden
from cbra.exceptions import InvalidAuthorizationScheme
from cbra.exceptions import UntrustedIssuer
from cbra.params import CurrentServer
from cbra.types import ICorsPolicy
from cbra.types import IPrincipal
from .const import GOOGLE_OPENID_SERVER
from .serviceaccountprincipal import ServiceAccountPrincipal


class GoogleEndpoint(Endpoint):
    """A :class:`cbra.Endpoint` implementation that is invoked by Google
    services, such as Cloud Scheduler or Pub/Sub.
    """
    __module__: str = 'cbra.ext.google'
    cors_policy: type[ICorsPolicy] = NullCorsPolicy
    method: str = "POST"
    principal: ServiceAccountPrincipal
    require_authentication: bool = True

    #: The codec configured with the Google JWKS used to verify the
    #: JWT.
    codec: PayloadCodec | None = None

    #: The set of subjects that are accepted by this endpoint.
    service_accounts: set[str] | None = None

    @classmethod
    async def principal_factory(
        cls,
        request: fastapi.Request,
        audience: str = CurrentServer,
        authorization: str | None = fastapi.Header(
            alias='Authorization',
            default=None,
            title="Authorization",
            description=(
                "The `Authorization` header must specify the `Bearer` scheme, "
                "holding a JWT that was signed by Google. The `iss` claim must "
                "be `https://accounts.google.com` and the signing keys to "
                "verify the signature are discovered using the OAuth 2.x/"
                "OpenID Connect metadata protocol. Any failure in fetching "
                "the JWKS, signature validation or JWT claims deserialization "
                "results in a `403` error response. If the `Authorization` "
                "header is missing, then this endpoint responds with a "
                "`401` status."
            )
        )
    ) -> IPrincipal | None:
        """Parse the ``Authorization`` header and verify that the bearer token
        was issued by Google and signed with the proper keys. Return the a new
        :class:`GooglePrincipal` instance.

        Args:
            authorization (str): the `Authorization` header specified on the
                HTTP request that is being authenticated.

        Returns:
            :class:`~cbra.ext.google.GooglePrincipal`

        Raises:
            :class:`~cbra.exceptions.AuthenticationRequired`: the `Authorization`
                header was not provided by the HTTP request and the endpoint class
                set the :attr:`require_authentication` to ``True``.
            :class:`~cbra.exceptions.InvalidAuthorizationScheme`: the `Authorization`
                header was present, but the scheme was invalid i.e. not `Bearer`.
            :class:`~cbra.exceptions.UntrustedIssuer`: the JWT was not issued
                by Google.
            :class:`~ckms.jose.Unusable`: the JWT payload could not be decoded,
                most probably because it was encrypted with a key that is not
                in the possession of the server.
            :class:`~ckms.jose.JOSEException`: any other issues with the JWT
                that were not listed above. Inspect the :attr:`JOSEException.code`
                attribute to determine the cause.
        """
        # If the token was not provided, then return None or raise an
        # error if authentication is required.
        if not authorization:
            if cls.require_authentication:
                raise AuthenticationRequired
            return None
    
        scheme, token = get_authorization_scheme_param(authorization or '')
        if str.lower(scheme or '') != 'bearer':
            raise InvalidAuthorizationScheme

        # Fetch the Google keys if they are not known yet, and verify the
        # signature
        if GoogleEndpoint.codec is None:
            async with httpx.AsyncClient() as client:
                metadata = await ServerMetadata.discover(
                    client=client,
                    issuer=GOOGLE_OPENID_SERVER
                )
                GoogleEndpoint.codec = PayloadCodec(
                    verifier=await metadata.get_jwks(client=client)
                )
        _, claims = await GoogleEndpoint.codec.jwt(token)
        if not isinstance(claims, JSONWebToken): # type: ignore
            raise MalformedPayload(http_status_code=403)
        assert claims.iss is not None # nosec
        if claims.iss != GOOGLE_OPENID_SERVER: # pragma: no cover
            # This should never happen, but check just to be sure.
            raise UntrustedIssuer(claims.iss)
        claims.verify(
            audience={
                audience,
                f'{request.url.scheme}://{request.url.netloc}{request.url.path}'
            },
            required={'iss', 'aud', 'sub', 'exp', 'email'}
        )
        assert claims.sub is not None # nosec
        assert claims.extra.get('email') is not None # nosec
        if cls.service_accounts is not None\
        and claims.extra['email'] not in cls.service_accounts:
            raise Forbidden(
                message=(
                    f"The service account or user {claims.extra['email']} is not "
                    "allowed to invoke this endpoint."
                )
            )
        return ServiceAccountPrincipal(
            email=claims.extra['email'],
            sub=str(claims.sub)
        )

    @Endpoint.register_exception # type: ignore
    async def on_uncaught_exception(
        self,
        exception: Exception
    ) -> fastapi.Response:
        """Ensures that the exception is suppresssed."""
        if not isinstance(self.transaction, fastapi.params.Depends):
            self.transaction.abort()
        self.logger.exception(
            "Caught fatal %s while handling message.",
            type(exception).__name__
        )
        return fastapi.Response(
            status_code=200,
            headers={'X-Error-Code': "UNCAUGHT_EXCEPTION"}
        )


# pylint: skip-file
from .browser import BrowserRenderer
from .json import IRenderer
from .json import JSONRenderer
from .null import NullRenderer
from .yaml import YAMLRenderer


__all__ = [
    'BrowserRenderer',
    'IRenderer',
    'JSONRenderer',
    'NullRenderer',
    'YAMLRenderer',
]